from time import time
import logging

logging.basicConfig()
logger = logging.getLogger('merkl')
LONG = False


def log_if_slow(f, msg, limit=0.25):
    t0 = time()
    ret = f()
    t1 = time()

    duration = t1 - t0
    if duration > limit:
        logger.debug(msg + f': {duration:.1f}s')

    return ret

def short_hash(hash):
    if hash is None or LONG:
        return hash

    return hash[:8]



import os
import math
import unittest
from pathlib import Path
from merkl import *
from merkl.exceptions import *
from merkl.tests import TestCaseWithMerklRepo
from merkl.io import FileRef, DirRef
from merkl.cache import get_cache_file_path, SqliteCache, BLOB_DB_SIZE_LIMIT_BYTES, MEMORY_CACHE
from merkl.utils import evaluate_futures, Eval
from merkl.util_tasks import combine_file_refs


class TestCache(TestCaseWithMerklRepo):
    def test_caching(self):
        task_has_run = False

        @task
        def task1(arg):
            nonlocal task_has_run
            task_has_run = True
            return arg

        val = task1('test').eval()
        self.assertEqual(val, 'test')
        # Make sure task_has_run was changed
        self.assertTrue(task_has_run)

        # Now reset and test that value is cached (task_has_run is not changed)
        task_has_run = False
        val = task1('test').eval()
        self.assertEqual(val, 'test')
        self.assertFalse(task_has_run)

    def test_file_outs(self):
        filename = '/tmp/merkl_test.txt'

        @task
        def my_task():
            with open(filename, 'w') as f:
                f.write('test')
            return FileRef(filename)

        out = my_task()
        out.eval()
        with open(get_cache_file_path(out.hash, 'txt'), 'r') as f:
            self.assertEqual(f.read(), 'test')

        self.assertTrue(os.path.exists(filename))
        # Check that file is removed with rm_after_caching=True is passed in

        @task
        def my_task():
            with open(filename, 'w') as f:
                f.write('test')
            return FileRef(filename, rm_after_caching=True)

        f = my_task().eval()
        self.assertFalse(os.path.exists(filename))
        self.assertNotEqual(f, filename)
        # Check that extension of file in cache is the same
        self.assertEqual(f.split('.')[-1], filename.split('.')[-1])

        self.assertTrue(my_task().in_cache())
        self.assertNotEqual(my_task().eval(), filename)
        self.assertEqual(my_task().eval().split('.')[-1], 'txt')

        # Check that temporary unnamed files work
        @task
        def my_task():
            global filename
            file_out = FileRef(ext='txt')
            filename = str(file_out)
            with open(file_out, 'w') as f:
                f.write('test')
            return file_out

        path = my_task().eval()
        self.assertTrue(path.endswith('.txt'))
        self.assertNotEqual(str(path), filename)
        self.assertFalse(os.path.exists(filename))  # temporary file should have been deleted

        with open(path, 'r') as f:
            self.assertEqual(f.read(), 'test')

        @task
        def my_task_single(arg):
            return arg, arg

        # Had a problem where both FileRefs from a batch task had same hash
        @batch(my_task_single)
        def my_task(args):
            files = []
            for arg in args:
                a = FileRef()
                with open(a, 'w') as f:
                    f.write('{arg} a')
                b = FileRef()
                with open(b, 'w') as f:
                    f.write('{arg} b')
                files.append((a, b))
            return files

        files = my_task(['arg1', 'arg2', 'arg3'])
        for file_a, file_b in files:
            self.assertNotEqual(file_a.hash, file_b.hash)
            file_a_out = file_a.eval()
            file_b_out = file_b.eval()
            self.assertNotEqual(file_a_out, file_b_out)

            # Make sure the files get deleted after clearing cache
            self.assertTrue(os.path.exists(file_a_out))
            file_a.clear_cache()
            self.assertFalse(os.path.exists(file_a_out))

    def test_dir_outs(self):
        filenames = []

        dir_path = None

        @task
        def my_task():
            nonlocal filenames, dir_path
            dir_out = DirRef()
            dir_path = str(dir_out)

            for i in range(5):
                filename = dir_out.get_new_file(ext='txt')
                filenames.append(filename)
                with open(filename, 'w') as f:
                    f.write(f'test{i}')

            return dir_out

        dir_out = my_task().eval()
        self.assertEqual(len(dir_out.files), 5)
        for i, (filename, orig_filename) in enumerate(zip(dir_out.files, filenames)):
            self.assertNotEqual(filename, orig_filename)
            with open(filename, 'r') as f:
                self.assertEqual(f.read(), f'test{i}')

        self.assertNotEqual(str(dir_out), dir_path)
        self.assertFalse(os.path.exists(dir_path))  # temporary dir should have been deleted

        # Test caching
        dir_out = my_task()
        self.assertTrue(dir_out.in_cache())
        dir_out = dir_out.eval()
        for i, filename in enumerate(dir_out.files):
            with open(filename, 'r') as f:
                self.assertEqual(f.read(), f'test{i}')

        # Test load_files()
        @task
        def my_task():
            dir_out = DirRef()
            for i in range(5):
                filename = str(Path(dir_out) / f'file{i}.txt')
                with open(filename, 'w') as f:
                    f.write(f'test{i}')

            dir_out.load_files()
            self.assertEqual(len(dir_out.files), 5)

            return dir_out

        my_task().eval()

        # Test combine_file_refs / DirRef.link_file_ref_into_dir()
        num_files = 5

        @task(outs=num_files)
        def my_task():
            refs = []
            for i in range(num_files):
                file_ref = FileRef()
                with open(file_ref, 'w') as f:
                    f.write(str(i))
                refs.append(file_ref)
            return refs

        file_refs = my_task()
        dir_ref = combine_file_refs(file_refs)
        dir_ref = dir_ref.eval()
        self.assertEqual(len(dir_ref.files), num_files)
        self.assertTrue(os.path.exists(file_refs[0].eval()))
        dir_file_path = Path(dir_ref) / file_refs[0].eval()
        self.assertTrue(os.path.exists(dir_file_path))
        with open(dir_file_path, 'r') as f:
            f1 = f.read()

        with open(file_refs[0].eval(), 'r') as f:
            f2 = f.read()

        self.assertEqual(f1, f2)

    def test_future_serialization(self):
        @task
        def my_task():
            return 3

        future = my_task()
        serialized = future.serializer.dumps(future)
        unserialized = future.serializer.loads(serialized)
        self.assertEqual(unserialized.serializer, future.serializer)

    def test_none_inputs(self):
        # This would break before when I did a manual check for value type serializability, leaving it here just in case
        @task
        def my_task(inpt=None):
            return inpt

        my_task().eval()

    def test_cache_temporarily(self):
        @task
        def my_task(a):
            return 2*a

        a = my_task(0)
        a.cache_temporarily = True
        a.eval()
        self.assertTrue(a.in_cache())
        b = my_task(a)
        b.eval()
        self.assertFalse(a.in_cache())
        self.assertTrue(b.in_cache())

    def test_cache_larger_blobs(self):
        # First test that small blob doesn't get stored in file
        hash = 'gfafasf323'
        SqliteCache.add(hash, b'small blob')
        self.assertFalse(os.path.exists(get_cache_file_path(hash)))

        # Then check that large blog is stored in file
        hash = 'gy25uhg2hwy34'
        big_blob = b'big blob!!'
        big_blob = big_blob * (math.ceil(BLOB_DB_SIZE_LIMIT_BYTES / len(big_blob)) + 1)  # make it large
        SqliteCache.add(hash, big_blob)
        self.assertTrue(os.path.exists(get_cache_file_path(hash)))

        # Check that file is removed when cache is cleared
        SqliteCache.clear(hash)
        self.assertFalse(os.path.exists(get_cache_file_path(hash)))

    def test_siblings_evaluated(self):
        # There's an issue where if a function has multiple outs, and only some of them have been evaluated before the
        # program crashes for some reason, then those outs that were not evaluated are never cached, even though the
        # result is in the shared cache. Therefore, changed it so that all sibling futures get evaluated as well, so
        # test that here
        @task
        def my_task():
            return 1, 2, 3

        @task
        def my_second_task(val):
            if val != 2:
                return val
            else:
                raise KeyboardInterrupt

        outs1 = my_task()
        outs2 = [my_second_task(out) for out in outs1]
        outs2[0].eval()
        with self.assertRaises(KeyboardInterrupt):
            outs2[1].eval()

        # Eval-ing out22 failed, so outs1[2] was never reached, but all outs from first stage should be cached
        for out in outs1:
            self.assertTrue(out.in_cache())

    def test_clear_all_except(self):
        @task
        def my_task(val):
            return 3, b'xxxxxxxxxx' * 20000  # something above the 100k limit for storing directly in DB

        out1, out2 = my_task(2)
        out1.eval()
        # Some sanity checks
        self.assertFalse(os.path.exists(get_cache_file_path(out1.hash)))
        self.assertTrue(os.path.exists(get_cache_file_path(out2.hash)))
        self.assertTrue(SqliteCache.has(out1.hash))
        self.assertTrue(SqliteCache.has(out2.hash))

        # Now let's say we evaluate my_task with some other input
        out1_2, out2_2 = my_task(3)
        out1_2.eval()
        self.assertTrue(SqliteCache.has(out2_2.hash))

        # Now clear all except this out
        SqliteCache.clear_all_except([out1_2.hash])

        # None of these should be cached or on file
        self.assertFalse(SqliteCache.has(out2_2.hash))
        self.assertFalse(SqliteCache.has(out1.hash))
        self.assertFalse(SqliteCache.has(out2.hash))
        self.assertFalse(os.path.exists(get_cache_file_path(out2.hash)))

        # Make sure pipeline outs remain (there was a bug with this)
        @pipeline
        def my_pipeline():
            return my_task(3)

        out1, out2 = my_pipeline()
        out1.eval()

        cache.clear([out1], keep=True)
        self.assertTrue(SqliteCache.has(out1.hash))
        self.assertTrue(SqliteCache.has(out1.parent_pipeline_future.hash))

    def test_pipeline_cache_invalidation(self):
        # Test that a cached pipeline out is removed/recalculated if it contains a future which has not been evaled/cached

        @task
        def my_task(val):
            return 2*val

        called = False

        @pipeline
        def my_pipeline():
            nonlocal called
            called = True
            return my_task(3)

        out = my_pipeline() # this should cache the Future output from my_task, but we never evaluate it
        self.assertTrue(called)

        called = False
        out = my_pipeline()
        self.assertTrue(called)

        out.eval()
        called = False
        out = my_pipeline()
        self.assertFalse(called)

    def test_memory_caching(self):
        @task(cache_in_memory=True)
        def my_task(val):
            return 2*val

        out = my_task(2)
        self.assertEqual(len(MEMORY_CACHE), 0)
        out.eval()
        self.assertEqual(len(MEMORY_CACHE), 1)
        self.assertEqual(MEMORY_CACHE[out.hash], 4)

    def test_stats(self):
        SqliteCache.add('h1', b'123', fn_name='module1.function1')
        SqliteCache.add('h2', b'12345', fn_name='module1.function1')

        SqliteCache.add('h3', b'12', fn_name='module1.function2')
        SqliteCache.add('h4', b'1234', fn_name='module1.function2')
        stats = SqliteCache.get_stats()
        self.assertEqual(stats[0], ('module1.function1', 2, 3+5))
        self.assertEqual(stats[1], ('module1.function2', 2, 2+4))

        self.assertEqual(SqliteCache.get_stats('module1.function1'), (2, 3+5))

        # Test that large blobs stored on the filesystem are counted correctly
        SqliteCache.add('h5', b'12'*BLOB_DB_SIZE_LIMIT_BYTES, fn_name='module1.function3')
        self.assertEqual(SqliteCache.get_stats('module1.function3'), (1, 2*BLOB_DB_SIZE_LIMIT_BYTES))

        # Test that FileRefs returned from tasks are counted correctly
        num_bytes = 32
        @task
        def my_task():
            file_ref = FileRef()
            with open(file_ref, 'w') as f:
                f.write('a'*num_bytes)
            return file_ref

        my_task().eval()
        self.assertEqual(SqliteCache.get_stats('test_cache.my_task'), (1, num_bytes))

    def test_batch_stats(self):
        @task
        def my_task(val):
            return 2*val

        @batch(my_task)
        def my_batch_task(vals):
            return [2*val for val in vals]

        with Eval():
            my_batch_task([1,2,3])

        stats = SqliteCache.get_stats()
        # Test that it's reported as coming from the single fn
        self.assertEqual(stats[0][0], 'test_cache.my_task')

if __name__ == '__main__':
    unittest.main()



import functools
import clize

from merkl.io import migrate_output_files
from merkl.utils import import_module_function
from merkl import cache


def migrate_futures_wrapper(f, glob):
    @functools.wraps(f)
    def _wrap(*args, **kwargs):
        outs = f(*args, **kwargs)
        migrate_output_files(outs, glob)

    return _wrap


class MigrateAPI:
    def migrate(self, glob, module_function):
        function = import_module_function(module_function)
        function = migrate_futures_wrapper(function, glob)
        clize.run(function, args=['merkl-migrate', *self.unknown_args])



import clize
from sigtools.specifiers import forwards_to_function

from merkl.future import Future, map_future_to_value
from merkl.utils import nested_map, nested_collect, import_module_function
from merkl.dot import print_dot_graph
from merkl import cache


class CacheAPI:
    def cache(self, module_function, clear=False):
        if module_function is not None:
            count, size = cache.SqliteCache.get_stats(module_function)
            print(f'Num entries: {count}')
            print(f'Total size: {size} bytes ({size / 10e6}M)')
        else:
            for module_function, count, size in sorted(cache.SqliteCache.get_stats(), key=lambda x: x[2]):
                print(f'{size/10e6:.2f}M\t{count}\t{module_function}')



import os
import json
import hashlib
from collections import defaultdict
from functools import cached_property, partial

import dill

import merkl.cache
from merkl.exceptions import *
from merkl.cache import get_modified_time, MEMORY_CACHE
from merkl.logger import logger, log_if_slow, short_hash
from merkl.io import write_track_file, write_future, FileRef, DirRef, get_merkl_file_hash
from merkl.utils import OPERATORS, nested_map, nested_collect, function_descriptive_name, DelayedKeyboardInterrupt, Eval

def map_to_hash(val):
    if isinstance(val, Future):
        return {'_hash': val.hash}
    return val


def map_future_to_value(val):
    if isinstance(val, Future):
        return val.eval()
    return val


def to_bytes_maybe(bytes_or_str):
    if isinstance(bytes_or_str, str):
        return bytes(bytes_or_str, 'utf-8')
    return bytes_or_str


def _code_args_serializer_default(obj, fn_name):
    logger.debug(f'Argument of type {type(obj)} for function {fn_name} is not JSON serializable, serializing with dill instead')
    return str(dill.dumps(obj))


FUTURE_STATE_EXCLUDED = ['bound_args', '_fn', 'single_fn', 'outs_shared_futures', '_parent_futures', '_val', 'on_completed']

deps_hash_cache = {}

class Future:
    __slots__ = [
        '_fn', 'single_fn', 'fn_code_hash', 'outs', 'out_name', 'deps', 'cache', 'serializer', 'bound_args',
        'outs_shared_cache', '_hash', '_deps_args_hash', '_deps_hash', '_args_hash', 'meta', 'is_input', 'output_files', 'is_pipeline',
        'parent_pipeline_future', 'invocation_id', 'task_id', 'batch_idx', 'cache_temporarily', 'outs_shared_futures',
        '_parent_futures', 'cache_in_memory', 'ignore_args', 'on_completed', '_val', '_fn_descriptive_name',
    ]

    def __init__(
        self,
        fn=None,
        fn_code_hash=None,
        outs=1,
        out_name=None,
        deps=None,
        cache=None,
        serializer=None,
        bound_args=None,
        outs_shared_cache=None,
        hash=None,
        meta=None,
        is_input=False,
        output_files=None,
        is_pipeline=False,
        invocation_id=-1,
        task_id=-1,
        batch_idx=None,
        cache_temporarily=False,
        cache_in_memory=False,
        ignore_args=None,
        single_fn=None,
    ):
        self._fn = fn
        self.single_fn = single_fn
        self.fn_code_hash = fn_code_hash
        self.outs = outs
        self.out_name = out_name
        self.deps = deps
        self.cache = cache
        self.serializer = serializer
        self.bound_args = bound_args

        self._hash = hash
        self._deps_args_hash = None
        self._args_hash = None
        self._deps_hash = None

        # Cache for the all outputs with the respect to a function and its args
        self.outs_shared_cache = outs_shared_cache if outs_shared_cache is not None else {}
        self.meta = meta
        self.is_input = is_input
        self.output_files = output_files
        if output_files is not None and len(output_files) > 0 and serializer is None:
            raise SerializationError(f'No serializer set for future {self}')

        self.is_pipeline = is_pipeline
        self.parent_pipeline_future = None
        self.invocation_id = invocation_id
        self.task_id = task_id
        self.batch_idx = batch_idx
        self.cache_temporarily = cache_temporarily
        self.cache_in_memory = cache_in_memory
        self.ignore_args = ignore_args
        self.outs_shared_futures = None
        self.on_completed = None
        self._parent_futures = None
        self._val = None
        self._fn_descriptive_name = None

    @property
    def fn(self):
        # If self.single_fn is set, then self._fn is the batch version.
        # Sometimes we want to access the batch one, but by default we use the
        # single function
        return self.single_fn or self._fn

    @property
    def deps_hash(self):
        if self._deps_hash:
            return self._deps_hash

        # We cannot cache the deps with only the function as key, because the function could have been
        # wrapped by @task multiple times
        fn_task_key = f'{self._fn}-{self.task_id}'
        if fn_task_key in deps_hash_cache:
            self._deps_hash = deps_hash_cache[fn_task_key]
            return self._deps_hash

        default = partial(_code_args_serializer_default, fn_name=self.fn_descriptive_name)

        hash_data = {
            'function_deps': self.deps or [],
        }

        logger.info(f'Hashing deps for {self.fn_descriptive_name}')
        m = hashlib.sha256()
        try:
            m.update(bytes(json.dumps(hash_data, sort_keys=True, default=default), 'utf-8'))
        except (TypeError, dill.PicklingError):
            raise SerializationError(f'Value in args {hash_data} not JSON or dill-serializable')
        self._deps_hash = m.hexdigest()
        deps_hash_cache[fn_task_key] = self._deps_hash
        return self._deps_hash

    @property
    def args_hash(self):
        if self._args_hash:
            return self._args_hash

        if not self.bound_args:
            return None

        default = partial(_code_args_serializer_default, fn_name=self.fn_descriptive_name)

        arguments = self.bound_args.arguments
        if self.ignore_args is not None:
            arguments = {name: arg for name, arg in arguments.items() if name not in self.ignore_args}

        hash_data = {
            'arguments': nested_map(arguments, map_to_hash, convert_tuples_to_lists=True),
        }

        m = hashlib.sha256()
        try:
            m.update(bytes(json.dumps(hash_data, sort_keys=True, default=default), 'utf-8'))
        except (TypeError, dill.PicklingError):
            raise SerializationError(f'Value in args {hash_data} not JSON or dill-serializable')
        self._args_hash = m.hexdigest()
        return self._args_hash

    @property
    def deps_args_hash(self):
        if self._deps_args_hash:
            return self._deps_args_hash

        if not self.bound_args:
            return None

        m = hashlib.sha256()
        m.update(bytes(self.deps_hash, 'utf-8'))
        m.update(bytes(self.args_hash, 'utf-8'))
        self._deps_args_hash = m.hexdigest()
        return self._deps_args_hash

    @property
    def hash(self):
        if self._hash:
            return self._hash

        m = hashlib.sha256()
        m.update(bytes(self.deps_args_hash, 'utf-8'))
        m.update(bytes(str(self.out_name), 'utf-8'))
        self._hash = m.hexdigest()
        return self._hash

    @property
    def parent_futures(self):
        if self._parent_futures is not None:
            return self._parent_futures

        if not self.bound_args:
            self._parent_futures = []
            return []

        is_future = lambda x: isinstance(x, Future)
        if self.batch_idx is not None:
            self._parent_futures = nested_collect(self.bound_args.args[0][self.batch_idx], is_future)
        else:
            self._parent_futures = (
                nested_collect(self.bound_args.args, is_future) +
                nested_collect(self.bound_args.kwargs, is_future)
            )
        return self._parent_futures

    def in_cache(self):
        # First check the in-memory cache
        if self.cache_in_memory and self.hash in MEMORY_CACHE:
            return True

        for output_file, write_merkl_file in self.output_files or []:
            if not write_merkl_file:
                continue

            try:
                output_file_merkl_hash = get_merkl_file_hash(output_file)
            except TypeError:
                logger.warning(f'Unable to deserialize .merkl file for {output_file}')
                output_file_merkl_hash = None

            if output_file_merkl_hash == self.hash:
                return True

        if self.cache is None:
            return False

        return self.cache.has(self.hash)

    def get_cache(self):
        # First check the in-memory cache
        if self.cache_in_memory and self.hash in MEMORY_CACHE:
            val = MEMORY_CACHE[self.hash]
            # We don't want to store both unserialized and serialized value in the memory cache, so we need to serialize
            # it if we need the serialized data. We only need this if we have to write it to file
            serialized = to_bytes_maybe(self.serializer.dumps(val)) if self.output_files is not None and len(self.output_files) > 0 else None
            return val, serialized

        if self.cache is None:
            return None, None

        if self.cache.has(self.hash):
            val = self.cache.get(self.hash)
            if self.is_input:
                # reading from source file, not serialized
                return val, val

            deserialized = log_if_slow(lambda: self.serializer.loads(val), f'Deserializing {self.fn_descriptive_name} out {self.hash} slow')
            return deserialized, val

        # Not in regular cache, so check the output files:
        for output_file, write_merkl_file in self.output_files or []:
            if not write_merkl_file:
                continue

            output_file_merkl_hash = get_merkl_file_hash(output_file)
            if output_file_merkl_hash == self.hash:
                with open(output_file, 'rb') as f:
                    val = f.read()

                try:
                    deserialized = log_if_slow(lambda: self.serializer.loads(val), f'Deserializing {self.fn_descriptive_name} out {self.hash} slow')
                except:
                    logger.error(f'Unable to deserialize {output_file}')
                    raise

                return deserialized, val

    def clear_cache(self, delete_output_files=False):
        self._val = None

        if self.cache_in_memory and self.hash in MEMORY_CACHE:
            del MEMORY_CACHE[self.hash]

        if self.cache is None:
            return

        self.cache.clear(self.hash)

        if delete_output_files:
            for output_file, write_merkl_file in self.output_files or []:
                if os.path.exists(output_file):
                    os.remove(output_file)
                merkl_file = output_file + '.merkl'
                if write_merkl_file and os.path.exists(merkl_file):
                    os.remove(merkl_file)

    def write_output_files(self, specific_out, specific_out_bytes):
        for path, write_merkl_file in self.output_files or []:
            # Check if output file is up to date
            modified = get_modified_time(path)
            up_to_date = False
            if self.cache is not None:
                md5_hash, merkl_hash, modified_time = self.cache.get_latest_file(path)
                up_to_date = modified_time == modified and merkl_hash == self.hash

            # If not up to date, serialize and write the new file
            if not up_to_date:
                if specific_out_bytes is None:
                    specific_out_bytes = log_if_slow(lambda: to_bytes_maybe(self.serializer.dumps(specific_out)), f'Serializing {self.fn_descriptive_name} out {self.hash} slow')

                write_track_file(path, specific_out_bytes, self, self.cache, write_merkl_file)

    def eval(self):
        if self._val is not None:
            return self._val

        if self.in_cache():
            if isinstance(self.out_name, int):
                if self.out_name <= 5:
                    logger.debug(f'{self.fn_descriptive_name}:{self.out_name} ({short_hash(self.hash)}) output was cached')
                    if self.out_name == 5:
                        logger.debug(f'And {self.outs - self.out_name} more...')

            specific_out, specific_out_bytes = self.get_cache()
            self.write_output_files(specific_out, specific_out_bytes)
        else:
            specific_out, specific_out_bytes = self._eval()

        if self.cache_in_memory:
            MEMORY_CACHE[self.hash] = specific_out

        self._val = specific_out

        if self.on_completed is not None:
            self.on_completed()

        return specific_out

    def _eval(self):
        specific_out = None
        specific_out_is_ref = False
        outputs = None
        called_function = False
        if self.deps_args_hash and self.deps_args_hash in self.outs_shared_cache:
            outputs = self.outs_shared_cache.get(self.deps_args_hash)
        else:
            evaluated_args = nested_map(self.bound_args.args, map_future_to_value) if self.bound_args else []
            evaluated_kwargs = nested_map(self.bound_args.kwargs, map_future_to_value) if self.bound_args else {}
            logger.debug(f'Calling {self.fn_descriptive_name} (out_name={self.out_name})')
            called_function = True

            # In case an Eval manager was used, we need to reset it so that any calls inside `fn` are not also
            # evaled immediately
            with Eval(False):
                outputs = self._fn(*evaluated_args, **evaluated_kwargs)

            if self.deps_args_hash:
                self.outs_shared_cache[self.deps_args_hash] = outputs

                if self.cache:
                    self.cache.no_commit = True  # for efficiency, commit only after all futures have been cached

                for future in self.outs_shared_futures or []:
                    if future.hash == self.hash:
                        continue
                    future._eval()

                if self.cache:
                    self.cache.no_commit = False
                    self.cache.commit()

        if isinstance(outputs, tuple) and len(outputs) != self.outs and self.outs != 1:
            raise TaskOutsError(f'Wrong number of outputs: {len(outputs)}. Expected {self.outs}')
        elif isinstance(outputs, dict) and self.outs != 1:
            if isinstance(self.outs, int):
                raise TaskOutsError(f'Outs was int: {self.outs}, but not 1, but output is dict')
            elif len(outputs) != len(self.outs):
                raise TaskOutsError(f'Wrong number of outputs: {len(outputs)}. Expected {self.outs}')

        if self.batch_idx is not None:
            outputs = outputs[self.batch_idx]
        if self.out_name is not None:
            outputs = outputs[self.out_name]

        specific_out = outputs
        deep_refs = nested_collect(
            specific_out,
            lambda out, lvl: (isinstance(out, FileRef) or isinstance(out, DirRef)) and lvl >= 1,
            include_level=True,
        )
        if len(deep_refs) > 0:
            raise ValueError('FileRef and DirRef cannot be deeply nested in a task out')

        if self.cache is not None:
            if isinstance(specific_out, FileRef) or isinstance(specific_out, DirRef):
                specific_out = self.cache.transfer_ref(specific_out, self.hash)
                specific_out_is_ref = True

        if self.is_pipeline:
            # Set the pipeline function on all output futures
            for future in nested_collect(outputs, lambda x: isinstance(x, Future)):
                future.parent_pipeline_future = self

        specific_out_bytes = None
        if not self.is_input:  # Futures from io should not be cached (but is read from cache)
            if self.cache is not None:
                specific_out_bytes = to_bytes_maybe(self.serializer.dumps(specific_out))

                if not self.cache.has(self.hash):  # gotta check again
                    with DelayedKeyboardInterrupt():
                        # Cache needs to be fully done, otherwise we might have added data to sqlite but not file
                        ref = (specific_out if specific_out_is_ref else None)
                        logger.debug(f'Caching {self.fn_descriptive_name} {short_hash(self.hash)} ref={ref}, len(content_bytes)={len(specific_out_bytes)}')
                        self.cache.add(
                            self.hash,
                            specific_out_bytes,
                            ref=ref,
                            fn_name=self.fn_descriptive_name,
                        )

                if called_function:  # Make sure we only clear parent futures once for all the output futures
                    for parent_future in self.parent_futures:
                        if parent_future.cache_temporarily:
                            parent_future.clear_cache()

            self.write_output_files(specific_out, specific_out_bytes)

        return specific_out, specific_out_bytes

    @property
    def fn_descriptive_name(self):
        if not hasattr(self, '_fn_descriptive_name') or self._fn_descriptive_name is None:
            self._fn_descriptive_name = function_descriptive_name(self.fn)
        return self._fn_descriptive_name

    def __repr__(self):
        return f'<Future: {self.hash[:8]}>'

    def __getstate__(self):
        # NOTE: a future is only pickled/serialized when it is returned by a pipeline, or when tracked in a file

        if self.cache is None:
            raise SerializationError(f'Serializing {repr(self)} ({self.fn_descriptive_name}: {self.out_name}) but there is no cache set')

        # Trigger calculation of _hash and _code_args_hash that we want to serialize
        self.hash

        # When we pickle a Future (for returning from pipelines), we don't want to pickle the whole graph and
        # potentially large data, so exclude e.g. `bound_args` which may contain futures
        # Also, the function may not be pickleable, and we don't need it when loading a cached Future
        state = {s: getattr(self, s, None) for s in self.__slots__ if s not in FUTURE_STATE_EXCLUDED}

        return state

    def __setstate__(self, d):
        for key, val in d.items():
            setattr(self, key, val)

        for key in FUTURE_STATE_EXCLUDED:
            setattr(self, key, None)

    def __or__(self, other):
        if not hasattr(other, 'is_merkl'):
            return NotImplemented

        return other(self)

    def __gt__(self, path):
        if not isinstance(path, str):
            return NotImplemented

        return write_future(self, path)

    def __rshift__(self, path):
        if not isinstance(path, str):
            return NotImplemented

        return write_future(self, path, write_merkl_file=True)

    def __hash__(self):
        return int(self.hash, 16)

    def __eq__(self, other):
        if not hasattr(other, 'hash'):
            return False

        return self.hash == other.hash

    def deny_access(self, *args, **kwargs):
        raise FutureAccessError

    @classmethod
    def from_file(cls, path, raise_file_not_found=True, replace_output_files=True):
        if not os.path.exists(path + '.merkl'):
            if raise_file_not_found:
                raise FileNotFoundError(path + '.merkl')
            else:
                return None

        with open(path + '.merkl', 'rb') as f:
            future = dill.load(f)
            if replace_output_files:
                future.output_files = [(path, True)]
            return future

    @property
    def v(self):
        # Short hand for getting the value
        return self.eval()


# Override all the operators of Future to raise a specific exception when used
for name in OPERATORS:
    setattr(Future, name, Future.deny_access)



from merkl.cli.cli import main
from merkl.task import task, batch, pipeline, Future, HashMode
from merkl.io import read_future, write_future, path_future, FileRef, DirRef, IdentitySerializer, WrappedSerializer, migrate_output_files
from merkl.util_tasks import combine_file_refs
from merkl.utils import Eval



import os
import uuid
import json
import dill
import glob
import shutil
import hashlib
import collections
from pathlib import Path
from functools import partial
from datetime import datetime
from tempfile import mkdtemp

import merkl
from merkl.utils import get_hash_memory_optimized, nested_collect, collect_dag_futures
from merkl.exceptions import SerializationError
from merkl.logger import logger


cwd = ''


def _get_file_content(path, flags):
    with open(path, 'r'+flags) as f:
        return f.read()


def _get_tmp_filename(suffix='', dir=None):
    dir = dir if dir is not None else merkl.cache.get_tmp_dir()
    path = None
    while path is None or os.path.exists(path):
        path = str(Path(dir) / f'{uuid.uuid4().hex}{suffix}')
    return path


def path_future(path):
    if not os.path.exists(path):
        raise FileNotFoundError(path)

    md5_hash = fetch_or_compute_md5(path, store=True)

    def _return_path():
        return path

    return merkl.future.Future(_return_path, hash=md5_hash, meta=path, is_input=True)


def read_future(path, flags=''):
    if not os.path.exists(path):
        raise FileNotFoundError(path)

    md5_hash = fetch_or_compute_md5(path, store=True)
    f = partial(_get_file_content, path=path, flags=flags)
    return merkl.future.Future(f, hash=md5_hash, meta=path, is_input=True)


def write_future(future, path, write_merkl_file=False):
    if future.output_files is None:
        future.output_files = []
    future.output_files.append((path, write_merkl_file))
    return future


def fetch_or_compute_md5(path, cache=merkl.cache.SqliteCache, store=True):
    if not os.path.exists(path):
        raise FileNotFoundError(path)

    modified = os.stat(path).st_mtime
    md5_hash, _ = cache.get_file_mod_hash(path, modified)
    if md5_hash is None:
        md5_hash = get_hash_memory_optimized(path, mode='md5')
        if store:
            cache.track_file(path, modified, md5_hash=md5_hash)

    return md5_hash


def fetch_or_compute_dir_md5(files, cache=merkl.cache.SqliteCache, store=True):
    h = hashlib.new('md5')
    for file_path in files:
        file_md5 = fetch_or_compute_dir_md5(file_path, cache, store)
        h.update(bytes(file_md5, 'utf-8'))

    return h.hexdigest()


def _write_merkl_file(path, future):
    with open(path + '.merkl', 'wb') as f:
        dill.dump(future, f)


def get_merkl_file_hash(path):
    if not os.path.exists(path + '.merkl'):
        return None

    try:
        with open(path + '.merkl', 'rb') as f:
            return dill.load(f).hash
    except:
        raise TypeError(f'Unable to deserialize .merkl file: {path}.merkl')


def write_track_file(path, content_bytes, future, cache=merkl.cache.SqliteCache, write_merkl_file=False):
    logger.debug(f'Writing to path: {path}')
    if isinstance(content_bytes, FileRef):
        shutil.copy(content_bytes, path)
    elif isinstance(content_bytes, DirRef):
        shutil.copytree(content_bytes, path)
    else:
        with open(path, 'wb') as f:
            f.write(content_bytes)

    # NOTE: we could get the md5 hash and store it, but not strictly necessary for
    # files that merkl has "created". The md5 is necessary though for files _read_ by merkl but produced elsewhere
    cache.track_file(path, merkl_hash=future.hash)
    if write_merkl_file:
        _write_merkl_file(path, future)


def migrate_output_files(outs, files_glob=None):
    """ Migrates the .merkl file hash to the new one, if .merkl file exists for output file """
    futures = nested_collect(outs, lambda x: isinstance(x, merkl.future.Future))

    dag_futures = set()
    for future in futures:
        collect_dag_futures(future, dag_futures, include_parent_pipelines=True)

    files = None
    if files_glob is not None:
        files = glob.glob(files_glob)
        files = set(files)

    migrated_files = []
    for future in dag_futures:
        for output_file, write_merkl_file in future.output_files or []:
            if files is not None and output_file not in files:
                continue

            if not write_merkl_file:
                continue


            if os.path.exists(output_file):
                current_hash = None
                if os.path.exists(output_file+'.merkl'):
                    with open(output_file+'.merkl', 'rb') as f:
                        try:
                            current_hash = dill.load(f).hash
                        except:
                            logger.warning(f'Unable to deserialize {output_file}.merkl content')

                    if future.hash == current_hash:
                        logger.info(f'{output_file} has up-to-date hash: {current_hash}')
                        continue

                logger.info(f'Migrating {output_file} to {future.hash}')
                _write_merkl_file(output_file, future)
                migrated_files.append(output_file)

    return migrated_files


class FileRef(str):
    __slots__ = ['rm_after_caching', '_hash']

    def __new__(cls, path=None, ext='bin', rm_after_caching=False):
        if path is None:
            suffix = None if ext is None else f'.{ext}'
            path = _get_tmp_filename(suffix=suffix)

        return str.__new__(cls, path)

    def __init__(self, path=None, ext='.bin', rm_after_caching=False):
        if path is None:
            self.rm_after_caching = True  # always remove temporary files
        else:
            self.rm_after_caching = rm_after_caching

        self._hash = None

    def __repr__(self):
        return f'<FileRef {self}>'

    def hash_repr(self):
        return f'<FileRef {self}: {self.hash}>'

    def remove(self):
        os.remove(self)

    @property
    def hash(self):
        if self._hash is not None:
            return self._hash

        self._hash = fetch_or_compute_md5(self)
        return self._hash


class DirRef(str):
    __slots__ = ['rm_after_caching', '_hash', '_files']

    def __new__(cls, path=None, rm_after_caching=False, files=None):
        if path is None:
            path = mkdtemp(dir=merkl.cache.get_tmp_dir())
            os.makedirs(path, exist_ok=True)

        return str.__new__(cls, path)


    def __init__(self, path=None, rm_after_caching=False, files=None):
        if path is None:
            self.rm_after_caching = True  # always remove temporary files
        else:
            self.rm_after_caching = rm_after_caching

        self._hash = None
        self._files = [] if files is None else files

    def get_new_file(self, name=None, ext=None):
        if name is not None:
            file_path = str(Path(self) / name)
        else:
            suffix = None if ext is None else f'.{ext}'
            file_path = _get_tmp_filename(suffix=suffix, dir=self)

        self._files.append(Path(file_path).name)
        return file_path

    @property
    def files(self):
        return [str(Path(self) / name) for name in self._files]

    def load_files(self):
        """ Load filenames from the file system """
        self._files = os.listdir(self)

    def __repr__(self):
        return f'<DirRef {self}>'

    def hash_repr(self):
        return f'<DirRef {self}: {self.hash}>'

    def remove(self):
        shutil.rmtree(self)

    def add_file_ref(self, file_ref, link=True):
        filename = os.path.basename(file_ref)
        if link:
            os.link(file_ref, self.get_new_file(name=filename))
        else:
            shutil.copy(file_ref, self.get_new_file(name=filename))

    @property
    def hash(self):
        if self._hash is not None:
            return self._hash

        self._hash = fetch_or_compute_dir_md5(self.files)
        return self._hash


class IdentitySerializer:
    @classmethod
    def dumps(cls, val):
        return val

    @classmethod
    def loads(cls, val):
        return val


class WrappedSerializer:
    """ A way to wrap another serializer, for supplying args to it when dumping """

    def __init__(self, serializer, *dump_args, **dump_kwargs):
        self.serializer = serializer
        self.dump_args = dump_args
        self.dump_kwargs = dump_kwargs

    def dump(self, *args, **kwargs):
        return self.serializer.dump(*args, *self.dump_args, **kwargs, **self.dump_kwargs)

    def dumps(self, *args, **kwargs):
        return self.serializer.dumps(*args, *self.dump_args, **kwargs, **self.dump_kwargs)

    def load(self, *args, **kwargs):
        return self.serializer.load(*args, **kwargs)

    def loads(self, *args, **kwargs):
        return self.serializer.loads(*args, **kwargs)



import os
import sys
import logging
import argparse
import cProfile, pstats, io
from pstats import SortKey
from pathlib import Path

import merkl
from merkl.cli.init import InitAPI
from merkl.cli.run import RunAPI
from merkl.cli.migrate import MigrateAPI
from merkl.cli.dot import DotAPI
from merkl.cli.cache import CacheAPI
from merkl.logger import logger


class MerkLAPI:
    init = InitAPI()
    run = RunAPI()
    dot = DotAPI()
    cache = CacheAPI()
    migrate = MigrateAPI()


def main():
    cwd = Path.cwd()
    desc = 'MerkL CLI'

    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='Verbose output')
    parser.add_argument(
        '-l', '--long', action='store_true', help='Print long hashes for verbose output')
    parser.add_argument(
        '-p', '--profile', action='store_true', help='Profile the command')

    parser.set_defaults(command='help', subcommand='')
    subparsers = parser.add_subparsers()

    # ------------- HELP --------------
    help_parser = subparsers.add_parser('help', description='Display help')
    help_parser.set_defaults(command='help', subcommand='')

    version_parser = subparsers.add_parser('version', description='Display version')
    version_parser.set_defaults(command='version', subcommand='')

    # ------------- INIT --------------
    init_parser = subparsers.add_parser(
        'init', description='Initialize a .merkl cache for this project')
    init_parser.set_defaults(command='init', subcommand='init')

    # ------------- RUN --------------
    run_parser = subparsers.add_parser(
        'run', description='Run a task or pipeline function')
    run_parser.set_defaults(command='run', subcommand='run')
    run_parser.add_argument('-n', '--no-cache', action='store_true', help='Disable caching')
    run_parser.add_argument('-c', '--clear', action='store_true', help='Clear cache of any unrelated items before running')
    run_parser.add_argument('module_function', help='Module function to run (<module>.<function>)')

    # ------------- DOT --------------
    dot_parser = subparsers.add_parser(
        'dot', description='Output the DAG of a task or pipeline as a dot file')
    dot_parser.set_defaults(command='dot', subcommand='dot')
    dot_parser.add_argument('--rankdir', choices=['TB', 'LR'], help='Value for the rankdir dot graph parameter')
    dot_parser.add_argument('--transparent-bg', action='store_true', help='Option to output transparent background')
    dot_parser.add_argument('-n', '--no-cache', action='store_true', help='Disable caching')
    dot_parser.add_argument('module_function', help='Module function to use (<module>.<function>)')

    # ------------- CACHE --------------
    cache_parser = subparsers.add_parser(
        'cache', description='Lists or clears cache')
    cache_parser.set_defaults(command='cache', subcommand='cache')
    cache_parser.add_argument('-c', '--clear', action='store_true', help='Clears the cache values')
    cache_parser.add_argument('module_function', nargs='?', default=None, help='Module function to list or clear (<module>.<function>)')

    # ------------- MIGRATE --------------
    migrate_parser = subparsers.add_parser(
        'migrate', description='Migrate output files for a pipeline')
    migrate_parser.set_defaults(command='migrate', subcommand='migrate')
    migrate_parser.add_argument('-g', '--glob', help='Glob for files to migrate')
    migrate_parser.add_argument('module_function', help='Module function to migrate (<module>.<function>)')


    args, unknown_args = parser.parse_known_args()
    kwargs = dict(args._get_kwargs())

    # Pop commands that should not go to the route
    kwargs.pop('command')
    kwargs.pop('subcommand')
    verbose = kwargs.pop('verbose')
    long_hashes = kwargs.pop('long')
    profile = kwargs.pop('profile')
    if verbose:
        logger.setLevel(logging.DEBUG)
        merkl.logger.LONG = long_hashes

    if args.command == 'help':
        print('Use the --help option for help')
        exit(0)
    elif args.command == 'version':
        #print(merkl.__version__)
        exit(0)

    api = MerkLAPI()
    api_route = getattr(api, args.command)
    api_route.unknown_args = unknown_args
    route = getattr(api_route, args.subcommand)

    if profile:
        with cProfile.Profile() as pr:
            route(**kwargs)

        s = io.StringIO()
        sortby = SortKey.CUMULATIVE
        ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
        ps.print_stats()
        print(s.getvalue())
    else:
        route(**kwargs)



from merkl import task
from merkl.io import DirRef, FileRef


@task
def combine_file_refs(file_refs, link=True):
    dir_ref = DirRef()
    for file_ref in file_refs:
        print(file_ref)
        try:
            dir_ref.add_file_ref(file_ref, link=link)
        except:
            # Duplicate?
            pass

    return dir_ref



import os
import sqlite3

from merkl.cache import get_db_path, SqliteCache

class InitAPI:
    def init(self):
        if os.path.exists(get_db_path()):
            print('Repository .merkl already exists')
            return

        SqliteCache.create_cache()



import os
import shutil
import sqlite3
from typing import NamedTuple

import merkl
from merkl.logger import logger, short_hash
from merkl.utils import collect_dag_futures, nested_collect, function_descriptive_name

MEMORY_CACHE = {}

NO_CACHE = False

BLOB_DB_SIZE_LIMIT_BYTES = 100000  # see link further down on page and blob sizes


def get_merkl_path():
    from merkl.io import cwd
    return f'{cwd}.merkl/'


def get_tmp_dir():
    return f'{get_merkl_path()}tmp/'


def get_db_path():
    return f'{get_merkl_path()}cache.sqlite3'


def get_cache_dir_path(hash=None):
    base = f'{get_merkl_path()}cache/'
    if hash is not None:
        return f'{base}{hash[:2]}/'
    return base


def get_cache_file_path(hash, ext='bin', makedirs=False):
    cache_dir = get_cache_dir_path(hash)
    if makedirs:
        os.makedirs(cache_dir, exist_ok=True)
    return f'{cache_dir}{hash}.{ext}'


def get_cache_out_dir_path(hash, makedirs=False):
    cache_dir = get_cache_dir_path(hash)
    if makedirs:
        os.makedirs(cache_dir, exist_ok=True)
    cache_dir = f'{cache_dir}{hash}/'
    return cache_dir


def get_modified_time(path):
    try:
        return os.stat(path).st_mtime
    except:
        return None


def clear(outs, keep=False, keep_outs=False):
    from merkl.future import Future
    futures = nested_collect(outs, lambda x: isinstance(x, Future))

    dag_futures = set()
    for future in futures:
        collect_dag_futures(future, dag_futures, include_parent_pipelines=True)

    if keep:
        future_hashes = [future.hash for future in dag_futures]
        SqliteCache.clear_all_except(future_hashes)
    elif keep_outs:
        outs_future_hashes = [future.hash for future in futures]
        SqliteCache.clear_all_except(outs_future_hashes)
    else:
        for future in dag_futures:
            future.clear_cache()


class SqliteCache:
    connection = None
    no_commit = False

    @classmethod
    def connect(cls):
        if cls.connection is None:
            try:
                cls.connection = sqlite3.connect(get_db_path())
            except sqlite3.OperationalError:
                if not os.path.exists(get_merkl_path()):
                    print(".merkl doesn't exist, did you run 'merkl init'?")
                    exit(1)
                raise
            cls.cursor = cls.connection.cursor()

    @classmethod
    def commit(cls):
        cls.connection.commit()

    @classmethod
    def create_cache(cls):
        if os.path.exists(get_db_path()):
            return

        logger.debug(f'Creating cache')
        os.makedirs(get_merkl_path(), exist_ok=True)
        os.makedirs(get_tmp_dir(), exist_ok=True)
        os.makedirs(get_cache_dir_path(), exist_ok=True)
        cls.connect()

        # Increase page size for faster BLOB performance:
        # https://www.sqlite.org/intern-v-extern-blob.html#:~:text=A%20database%20page%20size%20of,a%20separate%20file%20are%20faster.
        cls.cursor.execute("PRAGMA page_size = 16384;")

        cls.cursor.execute("""
            CREATE TABLE cache (
                hash CHARACTER(64) PRIMARY KEY,
                data BLOB,
                size INTEGER,
                ref_path TEXT,
                ref_is_dir BOOL,
                module_function TEXT NULL
            )
        """)

        cls.cursor.execute("""
            CREATE TABLE files (
                path TEXT,
                modified INTEGER,
                merkl_hash CHARACTER(64),
                md5_hash CHARACTER(64),
                PRIMARY KEY (path, modified)
            )
        """)

    @classmethod
    def transfer_ref(cls, ref, hash):
        """ Transfers a FileRef/DirRef from the original place in the file system to the merkl cache, and returns
        the new ref with the new path """
        logger.debug(f'Transferring {ref}')
        if isinstance(ref, merkl.io.FileRef):
            splits = ref.split('.')
            ext = None if len(splits) == 1 else splits[-1]
            cache_file_path = get_cache_file_path(hash, ext, makedirs=True)
            if os.path.exists(cache_file_path):
                os.remove(cache_file_path)

            if ref.rm_after_caching:
                os.rename(ref, cache_file_path)
            else:
                shutil.copy(ref, cache_file_path)

            new_file_out = merkl.io.FileRef(cache_file_path)
            return new_file_out
        elif isinstance(ref, merkl.io.DirRef):
            cache_dir_path = get_cache_out_dir_path(hash, makedirs=True)
            if ref.rm_after_caching:
                os.rename(ref, cache_dir_path)
            else:
                shutil.copytree(ref, cache_dir_path)

            ref = merkl.io.DirRef(cache_dir_path, files=ref._files)
            return ref

    @classmethod
    def add(cls, hash, content_bytes=None, ref=None, fn_name=None):
        content_len = len(content_bytes) if content_bytes is not None else 0
        if ref is not None:
            content_len = os.stat(ref).st_size

        ref_path = None if ref is None else str(ref)
        ref_is_dir = isinstance(ref, merkl.io.DirRef)

        if ref is None and content_len > BLOB_DB_SIZE_LIMIT_BYTES:
            # Faster to store data in a file
            cache_file_path = get_cache_file_path(hash, makedirs=True)
            with open(cache_file_path, 'wb') as f:
                f.write(content_bytes)
            content_bytes = None

        cls.connect()
        cls.cursor.execute("INSERT INTO cache VALUES (?, ?, ?, ?, ?, ?)", (hash, content_bytes, content_len, ref_path, ref_is_dir, fn_name))
        if not cls.no_commit:
            cls.connection.commit()

    @classmethod
    def clear(cls, hash):
        logger.debug(f'Clearing {short_hash(hash)}')
        cls.connect()
        result = list(cls.cursor.execute("SELECT ref_path, ref_is_dir, data FROM cache WHERE hash=?", (hash,)))
        if len(result) == 0:
            return

        ref_path, ref_is_dir, data = result[0]
        if ref_path is not None:
            if ref_is_dir:
                shutil.rmtree(ref_path)
            else:
                os.remove(ref_path)

        if data is None:
            # Data stored on file, remove it
            os.remove(get_cache_file_path(hash))

        cls.cursor.execute("DELETE FROM cache WHERE hash=?", (hash,))
        if not cls.no_commit:
            cls.connection.commit()

    @classmethod
    def clear_all_except(cls, hashes):
        num_items_in_cache = list(cls.cursor.execute('SELECT COUNT(*) FROM cache'))[0][0]
        num_deleted = num_items_in_cache - len(hashes)
        if num_deleted > 0:
            logger.warning(f'Deleting approximately {num_deleted} items from cache')

        quoted_hashes = [f'"{hash}"' for hash in hashes]
        hashes_with_files = list(
            cls.cursor.execute(f"SELECT hash, ref_path, ref_is_dir FROM cache WHERE hash NOT IN ({','.join(quoted_hashes)}) AND (data IS NULL OR ref_path IS NOT NULL)")
        )
        cls.cursor.execute(f"DELETE FROM cache WHERE hash NOT IN ({','.join(quoted_hashes)})", )

        if len(hashes_with_files) > 0:
            logger.warning(f'Deleting {len(hashes_with_files)} files or directories from cache')

        for hash, ref_path, ref_is_dir in hashes_with_files:
            if ref_path is not None:
                if ref_is_dir:
                    shutil.rmtree(ref_path)
                else:
                    os.remove(ref_path)
            else:
                os.remove(get_cache_file_path(hash))

    @classmethod
    def track_file(cls, path, modified=None, merkl_hash=None, md5_hash=None):
        logger.debug(f'Hashing file {path} modified={modified} merkl_hash={short_hash(merkl_hash)} md5_hash={short_hash(md5_hash)}')
        modified = modified or get_modified_time(path)
        cls.connect()
        cls.cursor.execute("INSERT INTO files VALUES (?, ?, ?, ?)", (path, modified, merkl_hash, md5_hash))
        if not cls.no_commit:
            cls.connection.commit()

    @classmethod
    def get_file_mod_hash(cls, path, modified=None):
        modified = modified or get_modified_time(path)
        cls.connect()
        result = cls.cursor.execute("SELECT md5_hash, merkl_hash FROM files WHERE path=? AND modified=?", (path, modified))
        result = list(result)
        if len(result) == 0:
            return None, None

        assert len(result) == 1
        return result[0]

    @classmethod
    def get_latest_file(cls, path):
        cls.connect()
        result = cls.cursor.execute("""
            SELECT md5_hash, merkl_hash, modified
            FROM files
            WHERE path=? ORDER BY modified DESC
        """, (path,))
        result = list(result)
        if len(result) == 0:
            return None, None, None
        return result[0]

    @classmethod
    def get(cls, hash):
        cls.connect()
        result = cls.cursor.execute("SELECT data FROM cache WHERE hash=?", (hash,))
        result = list(result)
        if len(result) == 0:
            return None

        data = result[0][0]
        if data is None:
            with open(get_cache_file_path(hash), 'rb') as f:
                return f.read()

        return data

    @classmethod
    def has(cls, hash):
        cls.connect()
        result = cls.cursor.execute("SELECT COUNT(*) FROM cache WHERE hash=?", (hash,))
        result = list(result)
        return result[0][0] > 0

    @classmethod
    def has_file(cls, hash):
        cls.connect()
        result = cls.cursor.execute("SELECT COUNT(*) FROM files WHERE md5_hash=?", (hash,))
        result = list(result)
        return result[0][0] > 0

    @classmethod
    def get_stats(cls, module_function=None):
        cls.connect()
        if module_function is not None:
            return list(cls.cursor.execute("SELECT COUNT(*), SUM(size) FROM cache WHERE module_function=?", (module_function,)))[0]
        else:
            return list(cls.cursor.execute("SELECT module_function, COUNT(*), SUM(size) FROM cache GROUP BY module_function"))

    @classmethod
    def clear_module_function(cls, module_function):
        cls.connect()
        result = list(cls.cursor.execute("DELETE FROM cache WHERE module_function=?", (module_function,)))
        return result



import clize
from sigtools.specifiers import forwards_to_function

from merkl.future import Future, map_future_to_value
from merkl.utils import nested_map, nested_collect, import_module_function
from merkl.dot import print_dot_graph
from merkl import cache


def print_graph_wrapper(f, rankdir, transparent_bg, no_cache):
    @forwards_to_function(f)
    def _wrapper(*args, **kwargs):
        orig, cache.NO_CACHE = cache.NO_CACHE, no_cache
        futures = nested_collect(f(*args, **kwargs), lambda x: isinstance(x, Future))
        print_dot_graph(futures, rankdir, transparent_bg)
        cache.NO_CACHE = orig

    return _wrapper


class DotAPI:
    def dot(self, module_function, rankdir, transparent_bg, no_cache):
        function = import_module_function(module_function)
        function = print_graph_wrapper(function, rankdir, transparent_bg, no_cache)
        clize.run(function, args=['merkl-dot', *self.unknown_args])



import merkl.io
from merkl.future import Future
from merkl.utils import nested_collect
from merkl.exceptions import FutureAccessError
from merkl import logger


MAX_LEN = 40
MAX_DEPS = 3


def print_dot_graph_nodes(futures, target_fn=None, printed=set()):
    # NOTE: This is very bad code, but probably not worth spending time on...
    for future in futures:
        is_cached_pipeline = future.parent_pipeline_future is not None and len(future.parent_futures) == 0
        node_future = future.parent_pipeline_future if is_cached_pipeline else future
        out_name = future.batch_idx if future.batch_idx is not None else future.out_name

        node_future_hash = node_future.hash
        if not logger.LONG:
            node_future_hash = node_future_hash[:6]
        node_id = f'{out_name}_{node_future_hash}_{node_future.invocation_id}'
        deps_args_hash = None
        if node_future.deps_args_hash:
            deps_args_hash = f'{node_future.deps_args_hash}_{node_future.invocation_id}'
        if not future.is_input and deps_args_hash not in printed:
            fn_name = f'{future.fn_descriptive_name}: {future.fn_code_hash[:4]}'
            if node_future.batch_idx is not None:
                fn_name = f'{future.fn_descriptive_name} (batch): {future.fn_code_hash[:4]}'
            if node_future.fn_code_hash not in printed:
                # Only print a function's deps once, in case of multiple invocations (list may be long)
                clamped = len(future.deps) > MAX_DEPS + 1
                deps = node_future.deps
                if clamped:
                    deps = deps[:MAX_DEPS]

                deps = '|'.join(name[:MAX_LEN] for name, *_ in deps)
                if clamped:
                    deps += f'|... +{len(node_future.deps)-MAX_DEPS}'

                label = fn_name
                if len(node_future.deps) > 0:
                    label = f'{{{label}|{deps} }}'

                printed.add(node_future.fn_code_hash)
            else:
                label = fn_name

            print(f'\t"fn_{deps_args_hash}" [shape=record, label="{label}"];')
            printed.add(deps_args_hash)

        args_str = ''
        if node_future.bound_args:
            for key, val in node_future.bound_args.arguments.items():
                if len(nested_collect(val, lambda x: isinstance(x, Future))) > 0:
                    continue

                if node_future.batch_idx is not None:
                    val = val[node_future.batch_idx]
                    val_str = f"'{val}'" if isinstance(val, str) else str(val)
                    args_str += val_str[:MAX_LEN] + '\n'
                else:
                    val_str = f"'{val}'" if isinstance(val, str) else str(val)
                    args_str += (f'{key}={val_str}')[:MAX_LEN] + '\n'

            args_str = args_str.strip()


        if args_str:
            args_id = deps_args_hash
            args_label = args_str
            if node_future.batch_idx is not None:
                args_id = f'{args_id}_{out_name}'
                args_label = f'{out_name}: {args_label}'
            print(f'\t"fn_{args_id}_args" [shape=plain, style=solid, label="{args_label}"];')
            if f'{args_id}_{deps_args_hash}' not in printed:
                print(f'\t"fn_{args_id}_args" -> "fn_{deps_args_hash}";')
                printed.add(f'{args_id}_{deps_args_hash}')

        if node_id not in printed:
            color = 'green' if future.in_cache() or future.is_input else 'red'
            output_files = future.output_files or []
            if future.is_input or len(output_files) > 0:
                # NOTE: in this case we store the file path in meta
                path = future.meta if future.is_input else '<br/>'.join(path for path, _ in output_files)
                shape = 'cylinder'
                node_label = f'{path}<br/>{future.hash[:4]}'
            else:
                shape = 'parallelogram'
                if out_name is None:
                    node_label = future.hash[:4]
                else:
                    node_label = f'{out_name}: {future.hash[:4]}'

            label = f"< <font color='{color}'>{node_label}</font> >"

            print(f'\t"out_{node_id}" [shape={shape}, style=dashed, label={label}];')
            if not future.is_input:
                print(f'\t"fn_{deps_args_hash}" -> "out_{node_id}"')

            printed.add(node_id)

        edge_name = f'{node_id}-{target_fn}'
        if target_fn and edge_name not in printed:
            print(f'\t"out_{node_id}" -> "fn_{target_fn}"')
            printed.add(edge_name)

        print_dot_graph_nodes(node_future.parent_futures, deps_args_hash, printed)


def print_dot_graph(futures, rankdir=None, transparent_bg=False):
    printed = set()
    print('digraph D {')
    if rankdir is not None:
        print(f'\trankdir="{rankdir}";')
    if transparent_bg:
        print('\tbgcolor="transparent";')
    print('\tnode [shape=plaintext, fillcolor="white", style=filled];')
    print_dot_graph_nodes(futures, printed=printed)
    print('}')



import os
import json
import pickle
import dill
import shutil
import unittest
from time import sleep
from pathlib import Path
from unittest.mock import patch

import merkl.io
from merkl import *
from merkl.tests import TestCaseWithMerklRepo
from merkl.utils import get_hash_memory_optimized
from merkl.io import migrate_output_files


class TestIO(TestCaseWithMerklRepo):
    tmp_file = '/tmp/tmpfile.txt'
    tmp_file2 = '/tmp/tmpfile2.txt'

    def setUp(self):
        super().setUp()
        with open(self.tmp_file, 'w') as f:
            f.write('hello world')

        with open(self.tmp_file2, 'w') as f:
            f.write('hej världen')

    def tearDown(self):
        super().tearDown()
        for path in [self.tmp_file, self.tmp_file + '.merkl', self.tmp_file2, self.tmp_file2 + '.merkl']:
            try:
                os.remove(path)
            except:
                pass

    def test_read_future(self):
        with self.assertRaises(FileNotFoundError):
            read_future('non_existant_file.txt', '')

        called = False
        def _mock(path, mode):
            nonlocal called
            called = True
            return get_hash_memory_optimized(path, mode)

        with patch('merkl.io.get_hash_memory_optimized', _mock):
            ff = read_future(self.tmp_file, '')
            self.assertEqual(ff.eval(), 'hello world')

        self.assertTrue(called)
        called = False

        # Make sure this expensive hashing doesn't happen again the second time
        with patch('merkl.io.get_hash_memory_optimized', _mock):
            ff = read_future(self.tmp_file, '')

        self.assertFalse(called)

    def test_write_future(self):
        @task
        def task1():
            return b'some data'

        out = write_future(task1(), self.tmp_file)
        out.eval()

        with open(self.tmp_file, 'rb') as f:
            self.assertEqual(f.read(), pickle.dumps(b'some data'))

    def test_write_future_pipe_syntax(self):
        @task
        def task1():
            return b'some data2'

        @task
        def identity(arg):
            return arg

        out = task1() | identity > self.tmp_file
        out.eval()

        with open(self.tmp_file, 'rb') as f:
            self.assertEqual(f.read(), pickle.dumps(b'some data2'))

        # Make types other than strings raises errors
        with self.assertRaises(TypeError):
            out = task1() > 3

    def test_write_future_with_merkl_file(self):
        @task
        def task1():
            return b'some data2'

        out = task1()
        out = out >> self.tmp_file
        out.eval()

        with open(self.tmp_file, 'rb') as f:
            self.assertEqual(f.read(), pickle.dumps(b'some data2'))

        with open(self.tmp_file + '.merkl', 'rb') as f:
            self.assertEqual(dill.load(f).hash, out.hash)

        # Make types other than strings raises errors
        with self.assertRaises(TypeError):
            out = task1() >> 3

    def test_output_files(self):
        @task
        def task1(data):
            return data

        out = task1(["data"])
        out_file = '/tmp/output_file.pickle'
        out >> out_file
        out.eval()

        self.assertTrue(os.path.exists(out_file))
        self.assertTrue(os.path.exists(out_file + '.merkl'))

        os.remove(out_file)
        os.remove(out_file + '.merkl')

        out = task1(["data"])
        out_file = '/tmp/output_file.pickle'
        out >> out_file
        out.eval()

        self.assertTrue(os.path.exists(out_file))
        self.assertTrue(os.path.exists(out_file + '.merkl'))
        with open(out_file+'.merkl', 'rb') as f:
            hash1 = dill.load(f).hash

        out = task1(["data2"])
        out_file = '/tmp/output_file.pickle'
        out >> out_file
        out.eval()

        with open(out_file+'.merkl', 'rb') as f:
            hash2 = dill.load(f).hash

        self.assertNotEqual(hash1, hash2)


    def test_path_future(self):
        @task
        def task1(path):
            return path

        fpath1 = path_future(self.tmp_file)
        self.assertTrue(isinstance(fpath1, Future))
        self.assertEqual(task1(fpath1).eval(), self.tmp_file)

        fpath2 = path_future(self.tmp_file)
        self.assertEqual(fpath1.hash, fpath2.hash)

        with open(self.tmp_file, 'w') as f:
            f.write('goodbye world')
        fpath3 = path_future(self.tmp_file)

        self.assertNotEqual(fpath3.hash, fpath2.hash)


    def test_output_file_migration(self):
        @task
        def task1(value):
            return f'some data {value}'

        def pipeline(value):
            out = task1(value)
            out = out >> self.tmp_file
            return out

        out = pipeline(1).eval()

        with open(self.tmp_file + '.merkl', 'rb') as f:
            hash1 = dill.load(f).hash

        out = pipeline(2)
        self.assertFalse(out.in_cache())

        migrate_output_files(out)

        with open(self.tmp_file + '.merkl', 'rb') as f:
            hash2 = dill.load(f).hash

        self.assertNotEqual(hash1, hash2)

        # Now test that the new migrated out file means that out is in cache
        self.assertTrue(out.in_cache())

        # Test migrating with glob
        out = pipeline(3)
        migrated_files = migrate_output_files(out, '/tmp/tmpfile.*')
        self.assertEqual(len(migrated_files), 1)

        out = pipeline(4)
        migrated_files = migrate_output_files(out, '/tmp/other_tmpfile.*')
        self.assertEqual(len(migrated_files), 0)

if __name__ == '__main__':
    unittest.main()



import os
import sys
import ast
import json
import inspect
import hashlib
import textwrap
import signal
from importlib import import_module
from functools import wraps, lru_cache
from inspect import isfunction, ismodule, getmodule
from typing import NamedTuple
from stdlib_list import stdlib_list
from sigtools.specifiers import forwards_to_function, signature
import merkl
from merkl import cache
from merkl.exceptions import TaskOutsError, EvalError
from merkl.logger import logger


DEBUG = False

BUILTIN_MODULES = stdlib_list()

OPERATORS = [
    '__bool__', '__not__', '__lt__', '__le__', '__ne__', '__ge__', 'truth', 'is_', 'is_not',
    '__abs__', '__add__', '__and__', '__floordiv__', '__index__', '__invert__', '__lshift__', '__mod__', '__mul__',
    '__matmul__', '__neg__', '__pos__', '__pow__', '__sub__', '__truediv__', '__xor__',
    '__concat__', '__contains__', 'countOf', '__delitem__', '__getitem__', 'indexOf', '__setitem__', 'length_hint',
    '__len__', '__iadd__', '__iand__', '__iconcat__', '__ifloordiv__', '__ilshift__', '__imod__', '__imul__',
    '__imatmul__', '__ior__', '__irshift__', '__ipow__', '__isub__', '__itruediv__', '__ixor__',
]


def doublewrap(f):
    """
    a decorator decorator, allowing the decorator to be used as:
    @decorator(with, arguments, and=kwargs)
    or
    @decorator
    """
    @forwards_to_function(f)
    def new_dec(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]) and not hasattr(args[0], 'type'):
            # Actual decorated function
            return f(args[0])
        else:
            # Decorator arguments
            return lambda realf: f(realf, *args, **kwargs)

    return new_dec


def nested_map(structure, map_function, convert_tuples_to_lists=False, include_level=False, curr_level=0):
    if isinstance(structure, tuple):
        new_tuple = [
            nested_map(s, map_function, convert_tuples_to_lists, include_level, curr_level+1)
            for s in structure
        ]
        if convert_tuples_to_lists:
            return new_tuple
        return tuple(new_tuple)
    elif isinstance(structure, list):
        return [
            nested_map(s, map_function, convert_tuples_to_lists, include_level, curr_level+1)
            for s in structure
        ]
    elif isinstance(structure, dict):
        return {
            key: nested_map(val, map_function, convert_tuples_to_lists, include_level, curr_level+1)
            for key, val in structure.items()
        }

    args = [structure]
    if include_level:
        args.append(curr_level)

    return map_function(*args)


def nested_apply(structure, apply_function, include_level=False, curr_level=0):
    if isinstance(structure, tuple) or isinstance(structure, list):
        for s in structure:
            nested_apply(s, apply_function, include_level, curr_level+1)
    elif isinstance(structure, dict):
        for key, val in structure.items():
            nested_apply(val, apply_function, include_level, curr_level+1)

    args = [structure]
    if include_level:
        args.append(curr_level)

    apply_function(*args)


def nested_collect(structure, collect_function, include_level=False):
    collected = []
    def _collect(*args):
        arg = args if len(args) > 1 else args[0]
        if collect_function(*args):
            collected.append(arg)

        return arg

    nested_apply(structure, _collect, include_level=include_level)
    return collected


class FunctionDep(NamedTuple):
    name: str
    value: object


def find_function_deps(f):
    from merkl.future import Future

    module = getmodule(f)
    try:
        dedented_source = textwrap.dedent(inspect.getsource(f))
    except:
        return []
    name_nodes = [node for node in ast.walk(ast.parse(dedented_source)) if isinstance(node, ast.Name)]
    deps = []
    seen = set()
    for node in name_nodes:
        if node.id in seen:
            continue

        seen.add(node.id)

        if merkl.__dict__.get(node.id):
            continue  # skip references to merkl stuff

        dep = module.__dict__.get(node.id)

        if dep is None:
            continue

        if isinstance(dep, Future):
            deps.append(FunctionDep(node.id, dep))
            continue

        if not isfunction(dep) and not ismodule(dep):
            # Functions and modules are special cases, but anything else
            # that is not serializable to json we skip
            dep = nested_map(dep, map_set_and_dict_to_list)
            try:
                json.dumps(dep)
            except TypeError:
                continue

        # Skip stuff defined in builtin modules
        dep_module = inspect.getmodule(dep)
        dep_module_name = dep_module.__name__ if dep_module is not None else None
        if dep_module_name == '__main__':
            # When module is being run directly, name is not representative, have to use this hack:
            dep_module_name = os.path.splitext(os.path.basename(dep_module.__file__))[0]

        if dep_module_name in BUILTIN_MODULES:
            continue

        deps.append(FunctionDep(node.id, dep))

    return deps


def get_hash_memory_optimized(f_path, mode='md5'):
    h = hashlib.new(mode)
    with open(f_path, 'rb') as file:
        block = file.read(512)
        while block:
            h.update(block)
            block = file.read(512)

    return h.hexdigest()


def get_return_nodes(node, collect):
    if isinstance(node, ast.FunctionDef):
        # Don't collect return nodes in nested functions
        return
    elif isinstance(node, ast.Return):
        collect.append(node)
    elif hasattr(node, 'iter_child_nodes'):
        for child_node in node.iter_child_nodes():
            get_return_nodes(child_node, collect)
    elif hasattr(node, 'body'):
        for child_node in node.body:
            get_return_nodes(child_node, collect)


def get_function_return_info(f):
    try:
        dedented_source = textwrap.dedent(inspect.getsource(f))
    except:
        return set(), set()

    function_ast = ast.parse(dedented_source).body[0]
    return_nodes = []
    for node in ast.iter_child_nodes(function_ast):
        get_return_nodes(node, return_nodes)

    return_types, num_returns = [], []
    for node in return_nodes:
        type_name = type(node.value).__name__
        if type_name == 'Tuple':
            num_returns.append(len(node.value.elts))
        elif type_name == 'Dict':
            try:
                num_returns.append(tuple([key.s for key in node.value.keys]))
            except AttributeError:
                raise TaskOutsError(f'Returned dict literal contains non-string-literal keys')
        else:
            # It it's not tuple, we treat it as a single value
            num_returns.append(1)

        return_types.append(type_name)

    return set(return_types), set(num_returns)


def evaluate_futures(outs, no_cache):
    from merkl.future import Future, map_future_to_value

    orig, cache.NO_CACHE = cache.NO_CACHE, no_cache
    ret = nested_map(outs, map_future_to_value)
    cache.NO_CACHE = orig
    return ret


def collect_dag_futures(future, out_futures, include_parent_pipelines=False):
    out_futures.add(future)
    if include_parent_pipelines and future.parent_pipeline_future is not None:
        out_futures.add(future.parent_pipeline_future)

    for parent_future in future.parent_futures:
        if parent_future in out_futures:
            continue
        collect_dag_futures(parent_future, out_futures)


def _dummy(*args, **kwargs):
    return

default_signature = signature(_dummy)

def signature_with_default(f):
    try:
        return signature(f)
    except ValueError:
        return default_signature


def signature_args_kwargs(f):
    sig = signature_with_default(f)
    args = [v.name for v in sig.parameters.values() if v.default == inspect._empty]
    kwargs = {v.name: v.default for v in sig.parameters.values() if v.default != inspect._empty}
    return args, kwargs


@lru_cache(maxsize=None)
def function_descriptive_name(fn, include_module=True):
    fn_name = fn.__name__ if hasattr(fn, '__name__') else str(fn)
    if not include_module:
        return fn_name

    module = inspect.getmodule(fn)
    module_name = module.__name__ if module is not None else 'unknown'
    return f'{module_name}.{fn_name}'


class DelayedKeyboardInterrupt():
    def __enter__(self):
        self.signal_received = False
        self.old_handler = signal.signal(signal.SIGINT, self.handler)

    def handler(self, sig, frame):
        self.signal_received = (sig, frame)
        logger.debug('SIGINT received. Delaying KeyboardInterrupt.')

    def __exit__(self, type, value, traceback):
        signal.signal(signal.SIGINT, self.old_handler)
        if self.signal_received:
            self.old_handler(*self.signal_received)


eval_immediately = False

class Eval():
    def __init__(self, set_val=True):
        self.set_val = set_val

    def __enter__(self, *args, **kwargs):
        global eval_immediately
        self.prev_val = eval_immediately
        eval_immediately = self.set_val  # normally True

    def __exit__(self, *args, **kwargs):
        global eval_immediately
        eval_immediately = self.prev_val


def import_module_function(module_function_str: str):
    module_name, function_name = module_function_str.rsplit('.', 1)

    cwd = os.getcwd()
    if cwd not in sys.path:
        sys.path.append(cwd)

    module = import_module(module_name)
    function = getattr(module, function_name)
    return function


def map_set_and_dict_to_list(item):
    if isinstance(item, set):
        return list(sorted(list(item)))
    elif isinstance(item, dict):
        return list(sorted(item.items(), key=lambda x: x[0]))
    return item



import sys
import json
import unittest
from io import StringIO

import merkl
from merkl.tests.tasks.embed_bert import embed_bert, embed_bert_large
from merkl.tests.tasks.embed_elmo import embed_elmo
from merkl.tests.tasks.find_deps_test import my_task
from merkl.tests import TestCaseWithMerklRepo
from merkl.future import Future
from merkl.task import task, batch, pipeline, HashMode
from merkl.exceptions import *
from merkl.utils import get_hash_memory_optimized, Eval
from merkl.io import FileRef, DirRef


def get_stderr(f):
    saved_stderr = sys.stderr
    try:
        err = StringIO()
        sys.stderr = err
        # Call the function
        out = f()
        err_output = err.getvalue().strip()
    finally:
        sys.stderr = saved_stderr

    return out, err_output



@task(deps=['test_dep'])
def my_dep_task_for_pipeline():
    return 3


@task(version='0.1')
def my_version_task():
    return 1


class TestTask(TestCaseWithMerklRepo):
    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def test_task_hashing(self):
        # Test that hash is the same every time
        self.assertEqual(embed_bert('sentence').hash, embed_bert('sentence').hash)

        # Test that hash is different for different inputs
        self.assertNotEqual(embed_bert('sentence').hash, embed_bert('another sentence').hash)

        # Test that hash is different with different function 
        self.assertNotEqual(embed_elmo('sentence').hash, embed_bert('sentence').hash)

        # Test that hash is different with different function in same file
        self.assertNotEqual(embed_bert('sentence').hash, embed_bert_large('sentence').hash)

        # Test that we raise the right exception when trying to pass something non-serializable to a task
        class MyClass:
            pass

        with self.assertRaises(SerializationError):
            embed_bert(MyClass()).hash

    def test_outs(self):
        with self.assertRaises(TaskOutsError):  # <= 0
            @task(outs=0)
            def _task_zero_outs(input_value):
                return input_value, 3

        with self.assertRaises(TaskOutsError):  # <= 0
            @task(outs=-1)
            def _task_negative_outs(input_value):
                return input_value, 3

        with self.assertRaises(TaskOutsError):  # bad outs type
            @task(outs=1.0)
            def _task_float_outs(input_value):
                return input_value, 3

        @task
        def _task1(input_value):
            return input_value, 3

        outs = _task1('test')
        # Single output by default
        self.assertEqual(len(outs), 2)
        for out in outs:
            self.assertTrue(isinstance(out, Future))

        with self.assertRaises(TaskOutsError):  # num returns mismatch
            @task
            def _task1(input_value):
                if input_value > 4:
                    return input_value, 2, 3
                return input_value, 3

        # Mismatching return in nested should be ok (should not raise exception)
        @task
        def _task1(input_value):
            def _nested_function():
                return input_value, 2, 3
            return input_value, 3

        # Single return value of other type than tuple should just produce a single Future (except dict)
        @task
        def _task1(input_value):
            return [1, 2*input_value]

        out = _task1('test')
        self.assertTrue(isinstance(out, Future))

        # Now set outs to 2, so we get two separate futures
        @task(outs=2)
        def _task2(input_value):
            return input_value, 3

        outs = _task2('test')
        self.assertEqual(len(outs), 2)
        self.assertNotEqual(outs[0].hash, outs[1].hash)

        # Test `outs` as a function
        @task(outs=lambda input_value, k: k)
        def _task3(input_value, k):
            return input_value, 3

        outs = _task3('test', 4)
        self.assertEqual(len(outs), 4)

        # Test that the wrong function signature fails
        with self.assertRaises(TaskOutsError):
            @task(outs=lambda inpoot_value, k: k)
            def _task4(input_value, k):
                return input_value, 3

        # A differing default value should not raise error
        @task(outs=lambda input_value, *args, k: k)
        def _task4(input_value, *args, k=3):
            return input_value, 3

        # Check that we get error if we return wrong number of outs
        @task(outs=3)
        def _task5(input_value):
            return input_value, 3

        with self.assertRaises(TaskOutsError):
            a, b, c = _task5('test')
            a.eval()

        # Unless it's one out
        @task(outs=1)
        def _task5(input_value):
            return input_value, 3

        _task5('test').eval()

        # Test dicts
        @task
        def _task6(input_value):
            return {'out1': 3, 'out2': input_value}

        out = _task6(5)
        self.assertTrue(isinstance(out, dict))
        self.assertEqual(len(out), 2)
        self.assertEqual(set(out.keys()), {'out1', 'out2'})
        self.assertEqual(out['out1'].eval(), 3)
        self.assertEqual(out['out2'].eval(), 5)

        # Test that dicts with keys that are not string literals fail
        with self.assertRaises(TaskOutsError):
            @task
            def _task7(input_value):
                key = 'out1'
                return {key: 3, 'out2': input_value}

        # Test that `outs` can be set to an iterable of string keys
        @task(outs=['out1', 'out2'])
        def _task7(input_value):
            outs =  {'out1': 3, 'out2': input_value}
            return outs

        outs = _task7(2)
        self.assertEqual(outs['out1'].eval(), 3)
        self.assertEqual(outs['out2'].eval(), 2)

        # Test that the above works if the keys are returned in a lambda
        @task(outs=lambda input_value: ['out1', 'out2'])
        def _task7(input_value):
            outs =  {'out1': 3, 'out2': input_value}
            return outs

        outs = _task7(2)
        self.assertEqual(outs['out1'].eval(), 3)
        self.assertEqual(outs['out2'].eval(), 2)

    def test_pipelines(self):
        @task(outs=lambda input_values: len(input_values))
        def _task1(input_values):
            return [val**2 for val in input_values]

        import math
        @task
        def _task2(input_value):
            return math.sqrt(input_value)

        vals = [1, 1, 1, 4, 5, 6]
        outs_stage1 = _task1(vals)
        outs_stage2 = [_task2(val) for val in outs_stage1]
        self.assertEqual([out.eval() for out in outs_stage2], vals)

        # Test that all hashes are different
        self.assertEqual(len(set(out.hash for out in outs_stage2)), len(vals))

    def test_hash_modes(self):
        def _task1(input_value):
            return input_value, 3

        with self.assertRaises(TypeError):
            _task1_file = task(hash_mode='not a hash mode')(_task1)

        _task1_file = task(hash_mode=HashMode.MODULE)(_task1)
        _task1_function = task(hash_mode=HashMode.FUNCTION)(_task1)
        for out1, out2 in zip(_task1_file('test'), _task1_function('test')):
            self.assertNotEqual(out1.hash, out2.hash)

        # Test that two identical functions in two different files (with slightly different content)
        # are the same if HashMode.FUNCTION is used
        from merkl.tests.tasks.identical_task1 import identical_task as identical_task1
        from merkl.tests.tasks.identical_task2 import identical_task as identical_task2
        self.assertEqual(identical_task1('test').hash, identical_task2('test').hash)

        def fn_dep1(arg1, arg2):
            return arg1 + arg2

        def fn_dep2(arg1, arg2):
            return arg1 + arg2

        from merkl.tests.tasks import embed_bert as embed_bert_module
        _task1_module_dep = task(deps=[embed_bert_module])(_task1)
        _task1_function_dep1 = task(deps=[fn_dep1])(_task1)
        _task1_function_dep2 = task(deps=[fn_dep2])(_task1)
        _task1_string_dep = task(deps=['val1'])(_task1)
        _task1_bytes_dep = task(deps=[b'val1'])(_task1)
        hashes = [
            _task1_module_dep('test')[0].hash,
            _task1_function_dep1('test')[0].hash,
            _task1_function_dep2('test')[0].hash,
            _task1_string_dep('test')[0].hash,
            _task1_bytes_dep('test')[0].hash,
        ]
        # Check that all hashes are unique, except the last two which are equal
        self.assertEqual(len(set(hashes[:3])), 3)
        self.assertEqual(len(set(hashes[3:])), 1)

        deps = [name for (name, dep) in my_task.deps]
        self.assertEqual(len(deps), 2+2)  # +2 is function name and code hash
        self.assertTrue('my_global_variable' in deps)
        self.assertTrue('_my_fn' in deps)

        # Test FileRef as dep
        filepath = '/tmp/my_file.txt'
        with open(filepath, 'w') as f:
            f.write('test')

        @task(deps=[FileRef(filepath)])
        def my_dep_task():
            return 1

        self.assertEqual(my_dep_task.deps[0], f'<FileRef /tmp/my_file.txt: {get_hash_memory_optimized(filepath)}>')

        # Test DirRef as dep
        filepath2 = '/tmp/my_file2.txt'
        with open(filepath2, 'w') as f:
            f.write('test2')

        @task(deps=[DirRef('/tmp/')])
        def my_dep_task():
            return 1

        self.assertTrue(my_dep_task.deps[0].startswith('<DirRef /tmp/:'))

        @pipeline
        def my_pipeline():
            return my_dep_task_for_pipeline()

        self.assertEqual(len(my_pipeline.deps), 2+2)  # +2 is function name and code hash
        self.assertEqual(my_pipeline.deps[1], 'test_dep')

    def test_future_operator_access(self):
        # Test that Future cannot be accessed by checking some operators
        future = embed_bert('sentence')

        set([future])  # this one is fine

        with self.assertRaises(FutureAccessError):
            # Can't be used as a truth value in statements
            if future:
                print('hello world')

        with self.assertRaises(FutureAccessError):
            # Can't be iterated over
            for x in future:
                print('hello world')

        with self.assertRaises(FutureAccessError):
            # Can't get a length from it
            print(len(future))

        with self.assertRaises(FutureAccessError):
            future += 1

    def test_batch_tasks(self):
        def fun(arg):
            return 3

        with self.assertRaises(BatchTaskError):
            # `fun` is not decorated
            @batch(fun)
            def fun2(args):
                pass

        @task
        def fun3(arg):
            return 1, 3

        with self.assertRaises(BatchTaskError):
            # batch function should only have an 'args' argument
            @batch(fun3)
            def fun4(args, other_arg):
                pass

        _embed_bert = embed_bert.orig_fn

        called = 0
        num_inputs = None

        @batch(embed_bert)
        def embed_bert_batch(args):
            nonlocal called, num_inputs
            called += 1
            num_inputs = len(args)
            return [_embed_bert(arg) for arg in args]

        with self.assertRaises(BatchTaskError):
            embed_bert_batch('test1')  # not a list

        batch_outs = embed_bert_batch(['test1', 'test2', 'test3', 'test1', 'test4', 'test5'])
        single_outs = [embed_bert(arg) for arg in ['test1', 'test2', 'test3', 'test1']]
        for single_out, batch_out in zip(single_outs[:-1], batch_outs[:-1]):
            self.assertEqual(single_out.hash, batch_out.hash)
            self.assertEqual(single_out.eval(), batch_out.eval())

        self.assertEqual(batch_outs[0].hash, batch_outs[-3].hash)
        # test1-3 should have been cached from the evaluation of the single function
        self.assertEqual(called, 0)
        self.assertEqual(num_inputs, None)

        # Now if we evaluate test5, it should be called
        batch_outs = embed_bert_batch(['test1', 'test2', 'test3', 'test1', 'test4', 'test5'])
        batch_outs[-1].eval()
        self.assertEqual(called, 1)
        self.assertEqual(num_inputs, 2)

        num_inputs = None
        called = 0
        batch_outs = embed_bert_batch(['test1', 'test2', 'test3', 'test1', 'test6'])
        for out in batch_outs:
            res = out.eval()

        # This tests is for a bug that happened in caching
        self.assertEqual(batch_outs[0].eval(), ['t', '1'])

        self.assertEqual(called, 1)
        self.assertEqual(num_inputs, 1)

        called = 0

        @task(outs=1)
        def single(inpt):
            raise NotImplementedError

        @batch(single)
        def embed_bert_without_single(args):
            nonlocal called
            called += 1
            return [_embed_bert(arg) for arg in args]

        batch_outs_without_single = embed_bert_without_single(['test1', 'test2', 'test3', 'test1'])
        for batch_out, batch_out_without_single in zip(batch_outs, batch_outs_without_single):
            self.assertNotEqual(batch_out.hash, batch_out_without_single.hash)
            self.assertEqual(batch_out_without_single.eval(), batch_out.eval())

        self.assertEqual(called, 1)

        @task
        def single_multiple_outs(arg1, arg2):
            return (arg1*2, arg2*3)

        @batch(single_multiple_outs)
        def batch_with_multiple_outs(args):
            return [(arg1*2, arg2*3) for arg1, arg2 in args]

        outs = batch_with_multiple_outs([(1,2), (2,3), (3,4), (1,2)])
        self.assertEqual(outs[0][0].hash, outs[-1][0].hash)
        self.assertEqual(outs[0][1].hash, outs[-1][1].hash)

        self.assertEqual(outs[0][0].eval(), outs[-1][0].eval())
        for out in outs:
            out[0].eval()
            out[1].eval()

        # Test caching
        outs = batch_with_multiple_outs([(1,2), (2,3), (3,4), (1,2)])
        for out in outs:
            self.assertTrue(out[0].in_cache())
            self.assertTrue(out[1].in_cache())

        self.assertEqual(outs[0][0].eval(), 2)
        self.assertEqual(outs[0][1].eval(), 6)

        @task
        def single_one_arg(arg):
            return (arg*2, arg2*3)

        @batch(single_one_arg)
        def batch_two_args(args):
            return [(arg1*2, arg2*3) for arg1, arg2 in args]

        with self.assertRaises(TypeError):
            batch_two_args([(1,2), (2,3)])

        # Test that two batch functions with the same single-fn does not have the same output hash
        @task
        def single(arg):
            return arg

        @task
        def batch_fn(args):
            return [single(arg) for arg in args]

        out1 = batch_fn(['test'])

        @task
        def single(arg):
            return arg

        @task
        def batch_fn(args):
            return [single(arg)+'2' for arg in args]

        out2 = batch_fn(['test'])
        self.assertNotEqual(out1.hash, out2.hash)

        #
        # Test batch with keyword args passed to each invocation
        #
        @task
        def single_task(arg1, arg2, some_option):
            if some_option is True:
                return arg1 * arg2
            else:
                return arg1 + arg2

        with self.assertRaises(BatchTaskError):
            # Should raise because some_option is not a kwarg in single_task
            @batch(single_task)
            def batch_task(args, some_option=True):
                pass

        @task
        def single_task(arg1, arg2, some_option=True):
            if some_option is True:
                return arg1 * arg2
            else:
                return arg1 + arg2

        with self.assertRaises(BatchTaskError):
            @batch(single_task)
            def batch_task(args, some_option_that_does_not_exist=True):
                pass

        with self.assertRaises(BatchTaskError):
            # Should raise because some_option default value differs from single_task
            @batch(single_task)
            def batch_task(args, some_option=False):
                pass

        @batch(single_task)
        def batch_task(args, some_option=True):
            out = []
            for arg1, arg2 in args:
                if some_option:
                    out.append(arg1*arg2)
                else:
                    out.append(arg1+arg2)
            return out

        with Eval():
            self.assertEqual(batch_task([(1, 1), (2, 2), (3, 3)], some_option=True), [1, 4, 9])
            self.assertEqual(batch_task([(1, 1), (2, 2), (3, 3)], some_option=False), [2, 4, 6])
            # NOTE: some_option should default to True here:
            self.assertEqual(batch_task([(1, 1), (2, 2), (3, 3)]), [1, 4, 9])

        @task
        def single_task(arg, some_option=True):
            if some_option is True:
                return arg1 * arg2
            else:
                return arg1 + arg2

        @batch(single_task)
        def batch_task(args, some_option=True):
            out = []
            for arg in args:
                if some_option:
                    out.append(arg*2)
                else:
                    out.append(arg+2)
            return out

        with Eval():
            self.assertEqual(batch_task([1, 2, 3], some_option=True), [2, 4, 6])
            # There was a problem where an exception was raised if not all kwargs were supplied
            asd = batch_task([1, 2, 3])
            self.assertEqual(batch_task([1, 2, 3]), [2, 4, 6])
            self.assertEqual(batch_task([1, 2, 3], some_option=False), [3, 4, 5])


    def test_pipelines(self):
        @task(cache=None)
        def my_task(k):
            return k*k

        with self.assertRaises(SerializationError):
            # my_task has no cache set, so caching the pipeline result
            # would not cache the nested Future result

            @pipeline
            def my_pipeline(i):
                return my_task(i)

            my_pipeline(3)

        @task
        def my_other_task(m):
            return m ** 2

        @pipeline
        def my_pipeline(i):
            return my_other_task(my_task(i))

        fut1 = my_pipeline(2)
        self.assertFalse(fut1.in_cache())
        self.assertEqual(fut1.eval(), 16)

        fut2 = my_pipeline(2)
        self.assertTrue(fut2.in_cache())
        self.assertEqual(fut2.hash, fut1.hash)
        self.assertEqual(fut2.parent_futures, [])

    def test_pipe_syntax(self):
        @task
        def task1():
            return 1

        @task
        def task2(arg):
            return 2*arg

        res1 = task1() | task2
        res2 = task2(task1())
        self.assertEqual(res1.hash, res2.hash)

    def test_version(self):
        with self.assertRaises(ValueError):
            @task(version=1)
            def my_task():
                return 1

        self.assertEqual(my_version_task().fn_code_hash, 'test_task.my_version_task-0.1')

    def test_non_json_serializable_args(self):
        @task
        def my_task(arg):
            return arg

        from uuid import uuid1
        import json
        with self.assertRaises(TypeError):
            json.dumps(uuid1())

        out = my_task(uuid1())
        out.eval()

    def test_eval_context_manager(self):
        @task
        def my_task(arg):
            return 2*arg, 3*arg

        @task
        def my_task2(arg):
            return 5*arg

        with Eval():
            out1, out2 = my_task(3)
            out3 = my_task2(out1)

            self.assertEqual(out1, 6)
            self.assertEqual(out2, 9)
            self.assertEqual(out3, 6*5)

        @task
        def my_task3(arg):
            raise NotImplementedError
            return 5*arg

        # Test batch task. Before fixing, immediately evaluated futures called the single_fn instead of batch
        @batch(my_task3)
        def my_batch(args):
            return [my_task.orig_fn(arg) for arg in args]

        with Eval():
            outs = my_batch([2, 3, 4])
            out2 = my_task(outs)

    def test_ignore_args(self):
        @task(ignore_args=['test2', 'kwargs'])
        def my_task(test, test2, *args, arg=3, **kwargs):
            return arg

        hash1 = my_task(1, 2, 3, arg=5, barf=6).hash
        hash2 = my_task(1, 3, 3, arg=5, barf=6).hash
        hash3 = my_task(1, 3, 3, arg=5, barf='hello').hash

        self.assertEqual(hash1, hash2)
        self.assertEqual(hash2, hash3)

    def test_nested_task(self):
        # Just to make sure this works as expected

        @task
        def inner_task(arg):
            return 2*arg

        @task
        def outer_task(arg):
            return inner_task(arg).eval()

        self.assertEqual(outer_task(3).eval(), 6)

    def test_future_from_file(self):
        @task(serializer=json)
        def my_task(asd):
            return 2*asd

        out = my_task(4)
        out >> '/tmp/my_file.json'
        out.eval()

        out2 = Future.from_file('/tmp/my_file.json')
        self.assertEqual(out.hash, out2.hash)
        self.assertEqual(out.eval(), out2.eval())

    def test_on_completed(self):
        message = None

        @task(serializer=json)
        def my_task(asd):
            nonlocal message
            message = 'hello world'
            return 2*asd

        f = my_task(4)
        def _handler():
            self.assertEqual(message, 'hello world')

        f.on_completed = _handler
        f.eval()


if __name__ == '__main__':
    unittest.main()



import os
import shutil
import unittest
import merkl
from merkl.cli.init import InitAPI
from merkl.cache import SqliteCache


class TestCaseWithMerklRepo(unittest.TestCase):
    def setUp(self):
        os.makedirs('/tmp/', exist_ok=True)
        merkl.io.cwd = '/tmp/'
        api = InitAPI()
        api.init()

    def tearDown(self):
        shutil.rmtree('/tmp/.merkl/')
        merkl.io.cwd = None
        SqliteCache.connection = None





import functools
import clize

from merkl.utils import evaluate_futures, import_module_function
from merkl.dot import print_dot_graph
from merkl import cache


def evaluate_futures_wrapper(f, no_cache, clear):
    @functools.wraps(f)
    def _wrap(*args, **kwargs):
        outs = f(*args, **kwargs)

        evaluated_outs = evaluate_futures(outs, no_cache)

        # Clear old outs after new have been calculated
        if clear:
            cache.clear(outs, keep=True)

        return evaluated_outs

    return _wrap


class RunAPI:
    def run(self, module_function, no_cache, clear):
        function = import_module_function(module_function)
        # Function output values may contain Futures, so wrap the function to evaluate them
        function = evaluate_futures_wrapper(function, no_cache, clear)
        clize.run(function, args=['merkl-run', *self.unknown_args], exit=False)



import json
import hashlib
import textwrap
import dill
from pathlib import Path
from enum import Enum
from functools import lru_cache
from inspect import getsource, isfunction, ismodule, getmodule
from sigtools.specifiers import forwards_to_function
import merkl
from merkl.utils import (
    doublewrap,
    nested_map,
    nested_collect,
    get_function_return_info,
    find_function_deps,
    FunctionDep,
    get_hash_memory_optimized,
    signature_with_default,
    signature_args_kwargs,
    function_descriptive_name,
    map_set_and_dict_to_list,
)
from merkl.logger import logger, short_hash
from merkl.future import Future
from merkl.exceptions import *
from merkl.cache import SqliteCache
from merkl.io import DirRef, FileRef
from merkl.utils import Eval

next_invocation_id = 0
next_task_id = 0

class HashMode(Enum):
    MODULE = 1
    FUNCTION = 2
    FIND_DEPS = 3


@lru_cache(maxsize=None)
def code_hash(f, is_module=False, version=None, hash_key=None):
    module = None
    if is_module:
        module = getmodule(f)
    code_obj = module if is_module else f
    if version is not None:
        if not isinstance(version, str):
            raise ValueError(f'Task version was not a string: {version}: {type(version)}')
        if module is None:
            module = getmodule(f)
        return f'{module.__name__}.{f.__name__}-{version}'
    elif hash_key is not None:
        if not isinstance(hash_key, str):
            raise ValueError(f'Task hash_key was not a string: {hash_key}: {type(hash_key)}')
        return hash_key

    try:
        code = textwrap.dedent(getsource(code_obj))
    except:
        code = code_obj.__name__
    m = hashlib.sha256()
    m.update(bytes(code, 'utf-8'))
    return m.hexdigest()


def resolve_serializer(serializer, out_name):
    if serializer is None:
        # NOTE: we use dill instead of pickle as default, because dill can serialize a reference to
        # itself, while pickle can't. This is a problem when we want to serialize a Future(e.g. when
        # caching pipeline results)
        return dill
    elif isinstance(serializer, dict):
        return serializer.get(out_name, dill)  # default to dill if key doesn't exist
    else:
        return serializer


def resolve_outs_and_return_type(outs, args, kwargs, return_type):
    resolved_outs = outs(*args, **kwargs) if callable(outs) else outs
    is_iterable = (
        isinstance(resolved_outs, tuple) or
        isinstance(resolved_outs, set) or
        isinstance(resolved_outs, list)
    )
    if is_iterable:
        return_type = 'Dict'

    validate_outs(resolved_outs, return_type=return_type)

    enumerated_outs = None
    if is_iterable:
        enumerated_outs = resolved_outs
    elif isinstance(resolved_outs, int):
        enumerated_outs = range(resolved_outs)
    else:
        raise ValueError

    return resolved_outs, enumerated_outs, return_type


def validate_outs(outs, sig=None, return_type=None):
    if isinstance(outs, int):
        if outs <= 0:
            raise TaskOutsError(f'Outs: {outs} is not >= 1')
    elif isinstance(outs, list) or isinstance(outs, tuple) or isinstance(outs, set):
        for key in outs:
            if not isinstance(key, str):
                raise TaskOutsError(f'Out key is not a string: {key}')
        if return_type != 'Dict' and not sig:
            raise TaskOutsError(f'Out is list/tuple/set of keys but return type is not dict: {return_type}')
    elif callable(outs) and sig is not None:
        outs_sig = signature_with_default(outs)
        if outs_sig.parameters.keys() != sig.parameters.keys():
            raise TaskOutsError(f'Outs function signature does not match task: {outs_sig} vs {sig}')
    elif return_type != 'Dict':
        raise TaskOutsError(f'Bad outs type: {outs}')


def validate_resolve_deps(deps):
    # Collect deps for tasks
    extra_deps = []
    for dep in deps:
        if (
            isinstance(dep, FunctionDep) and
            isfunction(dep.value) and
            hasattr(dep.value, 'is_merkl') and
            dep.value.type in ['task', 'batch']
        ):
            extra_deps += validate_resolve_deps(dep.value.deps)

    resolved_deps = []
    for i in range(len(deps)):
        name = None
        dep = deps[i]
        if isinstance(dep, FunctionDep):
            name = dep.name
            dep = dep.value

        if isfunction(dep):
            if hasattr(dep, 'is_merkl'):
                dep = dep.orig_fn
            try:
                dep = f'<Function {dep.__name__}: {code_hash(dep)}>'
            except OSError:  # source code not available
                logger.debug(f'No source code found for dep: {dep}')
                continue
        elif ismodule(dep):
            try:
                dep = f'<Module {dep.__name__}: {code_hash(dep, is_module=True)}>'
            except OSError:  # source code not available
                logger.debug(f'No source code found for dep: {dep}')
                continue
        elif isinstance(dep, bytes):
            dep = dep.decode('utf-8')
        elif isinstance(dep, Future):
            dep =  f'<Future {dep.fn_descriptive_name} {dep.hash}>'
        elif isinstance(dep, FileRef) or isinstance(dep, DirRef):
            dep = dep.hash_repr()
        elif not isinstance(dep, str):
            dep = nested_map(dep, map_set_and_dict_to_list)
            try:
                dep = json.dumps(dep)
            except TypeError:
                raise SerializationError(f'Task dependency {dep} not serializable to JSON')

        if name is not None:
            resolved_deps.append((name, dep))
        else:
            resolved_deps.append(dep)

    return resolved_deps + extra_deps


def eval_futures(obj):
    if isinstance(obj, Future):
        return obj.eval()
    return obj


@doublewrap
def batch(
    batch_fn,
    single_fn=None,
    hash_mode=HashMode.FIND_DEPS,
    deps=None,
    cache=SqliteCache,
    serializer=None,
    version=None,
    hash_key=None,
    cache_in_memory=None,
    ignore_args=None,
):
    deps = deps or []
    if single_fn is None:
        raise BatchTaskError(f"'single_fn' has to be supplied")

    batch_fn_name = function_descriptive_name(batch_fn)
    single_fn_name = function_descriptive_name(single_fn)

    if hasattr(single_fn, 'has_batch_fn'):
        raise BatchTaskError(f"Trying to create batch task for single function that has already been used")

    if not hasattr(single_fn, 'is_merkl') or single_fn.type != 'task':
        raise BatchTaskError(f'Function {single_fn_name} is not decorated as a task')

    if not isinstance(hash_mode, HashMode):
        raise TypeError(f'Unexpected HashMode value {hash_mode} for function {batch_fn_name}')

    batch_fn_sig = signature_with_default(batch_fn)
    batch_args, batch_kwargs = signature_args_kwargs(batch_fn)
    single_args, single_kwargs = signature_args_kwargs(single_fn)
    if len(batch_args) != 1:
        raise BatchTaskError(f'Batch function {batch_fn_name} must have exactly one input arg (but possibly more kwargs)')

    for kwarg, default in batch_kwargs.items():
        if kwarg not in single_kwargs:
            raise BatchTaskError(f'kwarg {kwarg} passed to batch function {batch_fn_name} that does not exist in single function\'s kwargs')
        single_default = single_kwargs[kwarg]
        if single_default != default:
            raise BatchTaskError(f'{batch_fn_name} kwarg {kwarg} default value ({default}) differs from single function\'s default value ({single_default})')


    # We create/resolve all deps and add them to `single_fn`, such that results
    # that come out of the single_fn and batch task have a hash that depends on
    # the implementation of both
    if hash_mode == HashMode.FIND_DEPS:
        deps += find_function_deps(batch_fn)
    batch_fn_code_hash = code_hash(batch_fn, hash_mode == HashMode.MODULE, version, hash_key)
    deps.append(('batch_function_code_hash', batch_fn_code_hash))
    deps.append(('batch_function_name', function_descriptive_name(batch_fn, include_module=False)))
    deps = validate_resolve_deps(deps)
    for dep in deps:
        single_fn.deps.append(dep)

    single_fn.has_batch_fn = True
    single_fn.orig_fn.has_batch_fn = True

    @forwards_to_function(batch_fn)
    def wrap(args, **kwargs):
        global next_invocation_id

        if callable(args):
            raise BatchTaskError(f'Function {args} is not decorated as a task')
        elif not isinstance(args, list):
            raise BatchTaskError(f'Batch args {args} is not a list')

        outs = []
        outs_shared_cache = {}
        futures = []
        non_cached_outs_args = []
        invocation_id = next_invocation_id

        deps_hash = None

        for i, args_tuple in enumerate(args):
            orig_args_tuple = args_tuple
            # Validate that args is a list of tuples, or single_fn has single input
            if not isinstance(args_tuple, tuple):
                # If single_fn has multiple parameters, then `args_tuple` has to be a tuple
                args_tuple = (args_tuple,)

            # In case `eval_immediately` is set (i.e. we're inside an Eval
            # context), we need to reset this temporarily here, since we don't
            # want to call `single_fn` yet.
            with Eval(False):
                out = single_fn(*args_tuple, **kwargs)

            any_out_not_cached = False
            cached_futures = []
            for future in nested_collect(out, lambda x: isinstance(x, Future)):
                if ignore_args is not None:
                    future.ignore_args = ignore_args  # must set this before calculating hash

                # Set the invocation id to the same for all futures
                future.invocation_id = invocation_id
                # deps_hash is the same for all output Futures, so cache it
                if deps_hash is not None:
                    future._deps_hash = deps_hash
                else:
                    deps_hash = future.deps_hash

                # Trigger calculation of the hash, which will be cached
                future.hash
                # Swap out the function for the batch version
                future.single_fn = future.fn
                future._fn = batch_fn
                # Swap out deps_args_hash to the batch_fn code hash, so
                # that the the value can be taken out of the shared cache
                # NOTE: this doesn't need to have deps and stuff, because it's not used for persistent caching, just to
                # Store the output temporarily
                future._deps_args_hash = batch_fn_code_hash

                if cache_in_memory is not None:
                    future.cache_in_memory = cache_in_memory

                # Override optional parameters
                if cache:
                    future.cache = cache if not merkl.cache.NO_CACHE else None
                if serializer:
                    future.serializer = resolve_serializer(serializer, future.out_name)

                if not future.in_cache():
                    any_out_not_cached = True
                else:
                    cached_futures.append(future)

            if any_out_not_cached:
                for future in cached_futures:
                    future.clear_cache()  # clear the previous cached value

                for future in nested_collect(out, lambda x: isinstance(x, Future)):
                    # Need to set the shared cache to be shared across all batch invocations
                    # NOTE: important only share this cache for outs that are not already in cache
                    future.outs_shared_cache = outs_shared_cache
                    futures.append(future)

                non_cached_outs_args.append((out, orig_args_tuple))

            outs.append(out)

        for future in futures:
            future.outs_shared_futures = futures

        # Swap out the args to the final list of batch args with non-cached results
        batch_bound_args = batch_fn_sig.bind([args for _, args in non_cached_outs_args], **kwargs)
        batch_bound_args.apply_defaults()
        for i, (out, _) in enumerate(non_cached_outs_args):
            for future in nested_collect(out, lambda x: isinstance(x, Future)):
                future.bound_args = batch_bound_args
                # Store the batch index from where the out should pick its results
                future.batch_idx = i

        next_invocation_id = invocation_id + 1

        if len(outs) > 1000:
            logger.debug(f'Batch task {batch_fn_name} has many outs ({len(outs)}), beware that hashing this many outs as args to another task may be slow')

        if merkl.utils.eval_immediately:
            outs = nested_map(outs, eval_futures)

        return outs

    wrap.is_merkl = True
    wrap.type = 'batch'
    wrap.outs = 1
    wrap.orig_fn = batch_fn
    wrap.deps = single_fn.deps
    return wrap


@doublewrap
def task(
    f,
    outs=None,
    hash_mode=HashMode.FIND_DEPS,
    deps=None,
    cache=SqliteCache,
    serializer=None,
    sig=None,
    version=None,
    hash_key=None,
    cache_in_memory=False,
    ignore_args=None,
):
    global next_task_id
    deps = deps or []
    ignore_args = ignore_args or []
    sig = sig if sig else signature_with_default(f)

    return_type = None
    if outs is not None:
        validate_outs(outs, sig)
    else:
        # Get num outs from AST if possible
        return_types, num_returns = get_function_return_info(f)
        if len(return_types) != 1 and 'Tuple' in return_types:
            raise TaskOutsError(f'Mismatch of number of return values in function return statements: {f}')
        elif len(num_returns) != 1:
            if len(num_returns) == 0:
                raise TaskOutsError(f'{f} has no return statement, cannot deduce number of outs')
            else:
                raise TaskOutsError(f'Mismatch of number of return values in function return statements: {f}')

        outs = num_returns.pop()
        return_type = return_types.pop()

    if not isinstance(hash_mode, HashMode):
        raise TypeError(f'Unexpected HashMode value {hash_mode} for function {f}')

    fn_code_hash = code_hash(f, hash_mode == HashMode.MODULE, version, hash_key)

    if hash_mode == HashMode.FIND_DEPS:
        deps += find_function_deps(f)

    deps = validate_resolve_deps(deps)

    deps.append(('function_code_hash', fn_code_hash))
    deps.append(('function_name', function_descriptive_name(f, include_module=False)))
    task_id = next_task_id

    @forwards_to_function(f)
    def wrap(*args, **kwargs):
        global next_invocation_id
        bound_args = sig.bind(*args, **kwargs)
        bound_args.apply_defaults()

        resolved_outs, enumerated_outs, resolved_return_type = resolve_outs_and_return_type(
            outs, args, kwargs, return_type
        )

        outputs = {}
        outs_shared_cache = {}
        is_single = resolved_outs == 1 and resolved_return_type not in ['Tuple', 'Dict']
        futures = []
        deps_hash = None
        for out_name in enumerated_outs:
            out_serializer = resolve_serializer(serializer, out_name)

            future = Future(
                f,
                fn_code_hash,
                resolved_outs,
                None if is_single else out_name,
                deps,
                cache if not merkl.cache.NO_CACHE else None,
                out_serializer,
                bound_args,
                outs_shared_cache,
                invocation_id=next_invocation_id,
                task_id=task_id,
                cache_in_memory=cache_in_memory,
                ignore_args=ignore_args,
            )
            # `deps_hash` triggers an expensive calculation, but it's the
            # same for all output futures, so we cache it and set manually
            if not hasattr(f, 'has_batch_fn'):
                # NOTE: if this is a single_fn for a batch task, we don't want to trigger this calculation, as
                # it would do it for all outputs from the batch task
                if deps_hash is not None:
                    future._deps_hash = deps_hash
                else:
                    deps_hash = future.deps_hash
            outputs[out_name] = future
            futures.append(future)

        for future in futures:
            future.outs_shared_futures = futures

        next_invocation_id += 1

        if merkl.utils.eval_immediately:
            outputs = nested_map(outputs, eval_futures)

        if is_single:
            return outputs[0]

        outputs = outputs if resolved_return_type == 'Dict' else tuple(outputs.values())

        if len(outputs) > 1000:
            logger.debug(f'Task {f} has many outs ({len(outputs)}), beware that hashing this many outs as args to another task may be slow')

        return outputs

    wrap.is_merkl = True
    wrap.type = 'task'
    wrap.outs = outs
    wrap.deps = deps
    wrap.orig_fn = f
    next_task_id += 1
    return wrap


@doublewrap
def pipeline(f, hash_mode=HashMode.FIND_DEPS, deps=None, cache=SqliteCache, cache_in_memory=False, ignore_args=None):
    deps = deps or []
    ignore_args = ignore_args or []
    sig = signature_with_default(f)

    if not isinstance(hash_mode, HashMode):
        raise TypeError(f'Unexpected HashMode value {hash_mode} for function {f}')

    pipeline_code_hash = code_hash(f, hash_mode == HashMode.MODULE)

    if hash_mode == HashMode.FIND_DEPS:
        deps += find_function_deps(f)

    deps = validate_resolve_deps(deps)

    @forwards_to_function(f)
    def wrap(*args, **kwargs):
        global next_invocation_id
        bound_args = sig.bind(*args, **kwargs)
        bound_args.apply_defaults()

        pipeline_future = Future(
            f,
            pipeline_code_hash,
            1,
            None,
            deps,
            cache if not merkl.cache.NO_CACHE else None,
            dill,
            bound_args,
            is_pipeline=True,
            invocation_id=next_invocation_id,
            cache_in_memory=cache_in_memory,
            ignore_args=ignore_args,
        )

        if pipeline_future.in_cache():
            # If a pipeline output was cached but the futures never evaluated
            # (e.g. due to a crash), then when we unserialize and try to use
            # them, merkl will crash. To prevent this, we check that all
            # futures are in cache, otherwise we remove this pipeline result
            # from the cache and re-evaluate
            outs = pipeline_future.eval()

            out_futures = nested_collect(outs, lambda x: isinstance(x, Future))
            out_future_not_cached = False
            for future in out_futures:
                if not future.in_cache():
                    out_future_not_cached = True
                    break

            if out_future_not_cached:
                logger.debug(f'Pipeline output for {function_descriptive_name(f)} ({short_hash(pipeline_future.hash)}) was cached, but output futures were not, so re-evaluating')
                pipeline_future.clear_cache()
                outs = pipeline_future.eval()

            num_futures = len(out_futures)
        else:
            outs = pipeline_future.eval()
            out_futures = nested_collect(outs, lambda x: isinstance(x, Future))

        next_invocation_id += 1

        if len(out_futures) > 1000:
            logger.debug(f'Pipeline {f} has many output futures ({len(out_futures)}), beware that hashing this many futures as args to another task may be slow')

        if merkl.utils.eval_immediately:
            outs = nested_map(outs, eval_futures)

        return outs

    wrap.is_merkl = True
    wrap.type = 'pipeline'
    wrap.outs = 1
    wrap.deps = deps
    wrap.orig_fn = f
    return wrap



import re
from setuptools import setup, find_packages

try: # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError: # for pip <= 9.0.3
    from pip.req import parse_requirements


long_description = open('README.md', 'r').read()

install_reqs = parse_requirements('requirements.txt', session=False)
reqs = list(install_reqs)
try:
    reqs = [str(ir.req) for ir in reqs]
except:
    reqs = [str(ir.requirement) for ir in reqs]

setup(
    name="merkl",
    packages=find_packages(),
    entry_points={"console_scripts": ['merkl = merkl:main']},
    version="0.4",
    description="MerkL",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Martin Pettersson",
    author_email="me@martindbp.com",
    url="https://github.com/martindbp/merkl",
    install_requires=reqs
)



class FutureAccessError(Exception):
    pass


class SerializationError(Exception):
    pass


class BatchTaskError(Exception):
    pass


class TaskOutsError(Exception):
    pass


class EvalError(Exception):
    pass
